from bloqade.ir import (
    rydberg,
    detuning,
    hyperfine,
    Sequence,
    Field,
    Pulse,
    Uniform,
    Linear,
    # ScaledLocations,
    LevelCoupling,
)
from bloqade.ir.control.sequence import NamedSequence
from bloqade.ir import Interval
import pytest
from bloqade import cast
from io import StringIO
from IPython.lib.pretty import PrettyPrinter as PP


def test_lvlcouple_base():
    lc = LevelCoupling()
    assert lc.children() == []


def test_lvlcouple_hf():
    lc = hyperfine

    assert lc.print_node() == "HyperfineLevelCoupling"
    assert str(lc) == "hyperfine"

    mystdout = StringIO()
    p = PP(mystdout)
    lc._repr_pretty_(p, 2)

    assert mystdout.getvalue() == "HyperfineLevelCoupling\n"


def test_lvlcouple_ryd():
    lc = rydberg

    assert lc.print_node() == "RydbergLevelCoupling"
    assert str(lc) == "rydberg"

    mystdout = StringIO()
    p = PP(mystdout)
    lc._repr_pretty_(p, 2)

    assert mystdout.getvalue() == "RydbergLevelCoupling\n"


def test_seqence():
    # seq empty
    seq = Sequence()
    assert seq.pulses == {}

    # seq non-lvlcoupling:
    with pytest.raises(TypeError):
        Sequence({"c": {}})

    # seq non-lvlcoupling:
    with pytest.raises(TypeError):
        Sequence({rydberg: 3.3})

    # seq full
    f = Field({Uniform: Linear(start=1.0, stop="x", duration=3.0)})
    ps = Pulse({detuning: f})
    seq_full = Sequence({rydberg: ps})

    assert seq_full.children() == {"RydbergLevelCoupling": ps}
    assert seq_full.print_node() == "Sequence"

    # named seq:
    named = NamedSequence(seq_full, "qq")

    assert named.children() == {"sequence": seq_full, "name": "qq"}
    assert named.print_node() == "NamedSequence"


def test_slice_sequence():
    f = Field({Uniform: Linear(start=1.0, stop=2.0, duration=3.0)})
    ps = Pulse({detuning: f})
    seq_full = Sequence({rydberg: ps})
    itvl = Interval(cast(0), cast(1.5))

    # slice:
    slc = seq_full[0:1.5]

    assert slc.children() == {"sequence": seq_full, "interval": itvl}
    assert slc.print_node() == "Slice"

    mystdout = StringIO()
    p = PP(mystdout)
    slc._repr_pretty_(p, 2)

    assert (
        mystdout.getvalue()
        == "Slice\n"
        + "├─ sequence\n"
        + "│  ⇒ Sequence\n"
        + "│    └─ RydbergLevelCoupling\n"
        + "│       ⇒ Pulse\n"
        + "│         └─ Detuning\n"
        + "│            ⇒ Field\n"
        + "⋮\n"
        + "└─ interval\n"
        + "   ⇒ Interval\n"
        + "     ├─ start\n"
        + "     │  ⇒ Literal: 0\n"
        + "     └─ stop\n"
        + "        ⇒ Literal: 1.5"
    )


def test_append_sequence():
    f = Field({Uniform: Linear(start=1.0, stop=2.0, duration=3.0)})
    ps = Pulse({detuning: f})
    seq_full = Sequence({rydberg: ps})

    app = seq_full.append(seq_full)

    assert app.children() == [seq_full, seq_full]
    assert app.print_node() == "Append"

    mystdout = StringIO()
    p = PP(mystdout)
    app._repr_pretty_(p, 2)

    assert (
        mystdout.getvalue()
        == "Append\n"
        + "├─ Sequence\n"
        + "│  └─ RydbergLevelCoupling\n"
        + "│     ⇒ Pulse\n"
        + "│       └─ Detuning\n"
        + "│          ⇒ Field\n"
        + "⋮\n"
        + "└─ Sequence\n"
        + "   └─ RydbergLevelCoupling\n"
        + "      ⇒ Pulse\n"
        + "        └─ Detuning\n"
        + "           ⇒ Field\n"
        + "⋮\n"
    )


"""
seq = Sequence(
    {
        rydberg: {
            detuning: {
                Uniform: Linear(start=1.0, stop="x", duration=3.0),
                ScaledLocations({1: 1.0, 2: 2.0}): Linear(
                    start=1.0, stop="x", duration=3.0
                ),
            },
        }
    }
)

print(seq)
print(seq.name("test"))
print(seq.append(seq))
print(seq[:0.5])
"""



from bloqade.builder.base import Builder
from bloqade.ir.control.sequence import Sequence
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from bloqade.builder.sequence_builder import SequenceBuilder


class ProgramStart(Builder):
    """
    ProgramStart is the base class for a starting/entry node for building a program.
    """

    @property
    def rydberg(self):
        """
        - Specify the Rydberg level coupling.
        - Possible Next:

            -> `...rydberg.rabi`
                :: address rabi term

            -> `...rydberg.detuning`
                :: address detuning field

        Examples:
            >>> node = bloqade.start.rydberg
            >>> type(node)
            <class 'bloqade.builder.coupling.Rydberg'>

            - Rydberg level coupling have two reachable field nodes:

                - detuning term (See also [`Detuning`][bloqade.builder.field.Detuning])
                - rabi term (See also [`Rabi`][bloqade.builder.field.Rabi])

            >>> ryd_detune = bloqade.start.rydberg.detuning
            >>> ryd_rabi = bloqade.start.rydberg.rabi

        See [`Rydberg`][bloqade.builder.coupling.Rydberg] for more details.
        """
        from bloqade.builder.coupling import Rydberg

        return Rydberg(self)

    @property
    def hyperfine(self):
        """
        - Specify the Hyperfile level coupling.
        - Possible Next:

            -> `...hyperfine.rabi`
                :: address rabi term

            -> `...hyperfine.detuning`
                :: address detuning field


        Examples:
            >>> node = bloqade.start.hyperfine
            >>> type(node)
            <class 'bloqade.builder.coupling.Hyperfine'>

            - Hyperfine level coupling have two reachable field nodes:

                - detuning term (See also [`Detuning`][bloqade.builder.field.Detuning])
                - rabi term (See also [`Rabi`][bloqade.builder.field.Rabi])

            >>> hyp_detune = bloqade.start.hyperfine.detuning
            >>> hyp_rabi = bloqade.start.hyperfine.rabi


        See [Hyperfine][bloqade.builder.coupling.Hyperfine] for more details.
        """
        from bloqade.builder.coupling import Hyperfine

        return Hyperfine(self)

    def apply(self, sequence: Sequence) -> "SequenceBuilder":
        """apply an existing pulse sequence to the program."""
        from bloqade.builder.sequence_builder import SequenceBuilder

        return SequenceBuilder(sequence, self)



from .base import AtomArrangement, LocationInfo
from typing import List, Tuple, Any, Union


class ListOfLocations(AtomArrangement):
    __match_args__ = ("location_list",)

    def __init__(self, location_list: List[Union[LocationInfo, Tuple[Any, Any]]] = []):
        self.location_list = []
        for ele in location_list:
            if isinstance(ele, LocationInfo):
                self.location_list.append(ele)
            else:
                self.location_list.append(LocationInfo(ele, True))

        if location_list:
            self.__n_atoms = len(self.location_list)
            self.__n_dims = len(self.location_list[0].position)
        else:
            self.__n_atoms = 0
            self.__n_dims = None

        super().__init__()

    @property
    def n_atoms(self):
        return self.__n_atoms

    @property
    def n_dims(self):
        return self.__n_dims

    def enumerate(self):
        return iter(self.location_list)

    def __iter__(self):
        return iter(self.location_list)


start = ListOfLocations()
"""
    - Program starting node
    - Possible Next <LevelCoupling>

        -> `start.rydberg`
            :: address rydberg level coupling

        -> `start.hyperfine`
            :: address hyperfine level coupling

    - Possible Next <AtomArragement>

        -> `start.add_locations(List[Tuple[int]])`
            :: add multiple atoms to current register

        -> `start.add_location(Tuple[int])`
            :: add atom to current register
"""



from bloqade.builder.base import Builder
from bloqade.builder.coupling import LevelCoupling, Rydberg, Hyperfine
from bloqade.builder.sequence_builder import SequenceBuilder
from bloqade.builder.field import Field, Detuning, RabiAmplitude, RabiPhase
from bloqade.builder.spatial import SpatialModulation, Location, Uniform, Var, Scale
from bloqade.builder.waveform import WaveformPrimitive, Slice, Record, Sample, Fn
from bloqade.builder.assign import Assign, BatchAssign
from bloqade.builder.flatten import Flatten
from bloqade.builder.parallelize import Parallelize, ParallelizeFlatten
from bloqade.builder.parse.stream import BuilderNode, BuilderStream

import bloqade.ir as ir
from itertools import repeat
from typing import TYPE_CHECKING, Tuple, Union, Dict, List, Optional, Set

if TYPE_CHECKING:
    from bloqade.ir.routine.params import Params, ParamType
    from bloqade.ir.analog_circuit import AnalogCircuit


class Parser:
    stream: Optional["BuilderStream"] = None
    vector_node_names: Set[str] = set()
    sequence: ir.Sequence = ir.Sequence()
    register: Union[ir.AtomArrangement, ir.ParallelRegister, None] = None
    batch_params: List[Dict[str, "ParamType"]] = [{}]
    static_params: Dict[str, "ParamType"] = {}
    order: Tuple[str, ...] = ()

    def reset(self, builder: Builder):
        self.stream = BuilderStream.create(builder)
        self.vector_node_names = set()
        self.sequence = ir.Sequence()
        self.register = None
        self.batch_params = [{}]
        self.static_params = {}
        self.order = ()

    def read_address(self, stream) -> Tuple[LevelCoupling, Field, BuilderNode]:
        spatial = stream.eat([Location, Uniform, Var], [Scale])
        curr = spatial

        if curr is None:
            return (None, None, None)

        while curr.next is not None:
            if not isinstance(curr.node, SpatialModulation):
                break
            curr = curr.next

        if type(spatial.node.__parent__) in [Detuning, RabiAmplitude, RabiPhase]:
            field = spatial.node.__parent__  # field is updated
            if type(field) in [RabiAmplitude, RabiPhase]:
                coupling = field.__parent__.__parent__  # skip Rabi
            else:
                coupling = field.__parent__

            # coupling not updated
            if type(coupling) not in [Rydberg, Hyperfine]:
                coupling = None
            return (coupling, field, spatial)
        else:  # only spatial is updated
            return (None, None, spatial)

    def read_waveform(self, head: BuilderNode) -> Tuple[ir.Waveform, BuilderNode]:
        curr = head
        waveform = None
        while curr is not None:
            match curr.node:
                case Slice(start, stop, _):
                    waveform = waveform[start:stop]
                case Record(name, _):
                    waveform = waveform.record(name)
                case Sample(dt, interpolation, Fn() as fn_node):
                    if interpolation is None:
                        if self.field_name == ir.rabi.phase:
                            interpolation = ir.Interpolation.Constant
                        else:
                            interpolation = ir.Interpolation.Linear

                    sample_waveform = ir.Sample(
                        fn_node.__bloqade_ir__(), interpolation, dt
                    )
                    if waveform is None:
                        waveform = sample_waveform
                    else:
                        waveform = waveform.append(sample_waveform)

                case Fn() if curr.next is not None and isinstance(
                    curr.next.node, Sample
                ):  # skip this for the sample node above
                    pass
                case WaveformPrimitive() as wf:
                    if waveform is None:
                        waveform = wf.__bloqade_ir__()
                    else:
                        waveform = waveform.append(wf.__bloqade_ir__())
                case _:
                    break
            curr = curr.next

        return waveform, curr

    def read_spatial_modulation(
        self, head: BuilderNode
    ) -> Tuple[ir.SpatialModulation, BuilderNode]:
        curr = head
        spatial_modulation = None
        scaled_locations = ir.ScaledLocations({})

        while curr is not None:
            match curr.node:
                case Location(label, _):
                    scaled_locations[ir.Location(label)] = 1.0
                case Scale(value, Location(label, _)):
                    scaled_locations[ir.Location(label)] = ir.cast(value)
                case Uniform(_):
                    spatial_modulation = ir.Uniform
                case Var(name, _):
                    spatial_modulation = ir.RunTimeVector(name)
                    self.vector_node_names.add(name)
                case _:
                    break
            curr = curr.next

        if scaled_locations:
            return scaled_locations, curr
        else:
            return spatial_modulation, curr

    def read_field(self, head) -> ir.Field:
        sm, curr = self.read_spatial_modulation(head)
        wf, _ = self.read_waveform(curr)

        if wf is None or sm is None:
            return ir.Field({})

        return ir.Field({sm: wf})

    def read_sequeence(self) -> ir.Sequence:
        if isinstance(self.stream.curr.node, SequenceBuilder):
            # case with sequence builder object.
            self.sequence = self.stream.read().node.sequence
            return self.sequence

        stream = self.stream.copy()
        while stream.curr is not None:
            coupling_builder, field_builder, spatial_head = self.read_address(stream)
            if coupling_builder is not None:
                # update to new pulse coupling
                self.coupling_name = coupling_builder.__bloqade_ir__()

            if field_builder is not None:
                # update to new field coupling
                self.field_name = field_builder.__bloqade_ir__()

            if spatial_head is None:
                break

            pulse = self.sequence.pulses.get(self.coupling_name, ir.Pulse({}))
            field = pulse.fields.get(self.field_name, ir.Field({}))

            new_field = self.read_field(spatial_head)
            field = field.add(new_field)

            pulse.fields[self.field_name] = field
            self.sequence.pulses[self.coupling_name] = pulse

        return self.sequence

    def read_register(self) -> ir.AtomArrangement:
        # register is always head of the stream
        register_node = self.stream.read()
        self.register = register_node.node

        return self.register

    def read_pragmas(self) -> None:
        pragma_types = (
            Assign,
            BatchAssign,
            Flatten,
            Parallelize,
            ParallelizeFlatten,
        )

        stream = self.stream.copy()
        curr = stream.read_next(pragma_types)

        while curr is not None:
            match curr.node:
                case Assign(static_params):
                    self.static_params = {
                        name: value for name, value in static_params.items()
                    }

                case BatchAssign(batch_param):
                    tuple_iterators = [
                        zip(repeat(name), values)
                        for name, values in batch_param.items()
                    ]
                    self.batch_params = list(map(dict, zip(*tuple_iterators)))
                case Flatten(order):
                    self.order = order

                    seen = set()
                    dup = []
                    for x in order:
                        if x not in seen:
                            seen.add(x)
                        else:
                            dup.append(x)

                    if dup:
                        raise ValueError(f"Cannot flatten duplicate names {dup}.")

                    order_names = set([*order])
                    flattened_vector_names = order_names.intersection(
                        self.vector_node_names
                    )

                    if flattened_vector_names:
                        raise ValueError(
                            f"Cannot flatten RunTimeVectors: {flattened_vector_names}."
                        )

                case Parallelize(cluster_spacing) | ParallelizeFlatten(cluster_spacing):
                    self.register = ir.ParallelRegister(self.register, cluster_spacing)
                case _:
                    break

            curr = curr.next

    def parse_register(
        self, builder: Builder
    ) -> Union[ir.AtomArrangement, ir.ParallelRegister]:
        self.reset(builder)
        self.read_register()
        self.read_pragmas()
        return self.register

    def parse_sequence(self, builder: Builder) -> ir.Sequence:
        self.reset(builder)
        self.read_sequeence()
        return self.sequence

    def parse_circuit(self, builder: Builder) -> "AnalogCircuit":
        from bloqade.ir.analog_circuit import AnalogCircuit

        self.reset(builder)
        self.read_register()
        self.read_sequeence()
        self.read_pragmas()

        circuit = AnalogCircuit(self.register, self.sequence)

        return circuit

    def parse_source(self, builder: Builder) -> Tuple["AnalogCircuit", "Params"]:
        from bloqade.ir.routine.params import Params
        from bloqade.ir.analog_circuit import AnalogCircuit

        self.reset(builder)
        self.read_register()
        self.read_sequeence()
        self.read_pragmas()

        params = Params(
            static_params=self.static_params,
            batch_params=self.batch_params,
            flatten_params=self.order,
        )
        circuit = AnalogCircuit(self.register, self.sequence)

        return circuit, params



from typing import Optional, Union, List
from numbers import Real

from bloqade.builder.parse.trait import CompileJSON, Parse

ParamType = Union[Real, List[Real]]


class Builder(CompileJSON, Parse):
    __match_args__ = ("__parent__",)

    def __init__(
        self,
        parent: Optional["Builder"] = None,
    ) -> None:
        self.__parent__ = parent



from bloqade.builder.base import Builder


class BraketService(Builder):
    @property
    def braket(self):
        """
        - Specify braket service
        - Possible Next:

            -> `...braket.aquila`
                :: Aquila QPU, via braket service

            -> `...braket.local_emulator`
                :: braket local emulator backend
        """
        return BraketDeviceRoute(self)


class BraketDeviceRoute(Builder):
    def aquila(self):
        """
        Specify QuEra's Aquila QPU

        Return:
            BraketHardwareRoutine

        - Possible Next:

            -> `...aquila().submit`
                :: submit aync remote job

            -> `...aquila().run`
                :: submit job and wait until job finished
                and results returned

            -> `...aquila().__callable__`
                :: submit job and wait until job finished
                and results returned

        """
        from bloqade.ir.routine.base import Routine

        return Routine(self).braket.aquila()

    def local_emulator(self):
        """
        Using Braket local emulator

        Return:
            BraketLocalEmulatorRoutine


        - Possible Next:

            -> `...local_emulator().run`
                :: run on local emulator

        """
        from bloqade.ir.routine.base import Routine

        return Routine(self).braket.local_emulator()



# import bloqade.lattice as lattice

# geo = lattice.Square(3)
# prog = geo.rydberg.detuning
# for i in range(5):
#     prog = prog.location(i)
# prog.linear(start=1.0, stop=2.0, duration="x")
# import pytest
import bloqade.ir as ir
from bloqade.builder import spatial
from bloqade.builder import waveform

# import bloqade.builder.backend as builder_backend
import bloqade.ir.routine.quera as quera
import bloqade.ir.routine.braket as braket


from bloqade.ir import rydberg, detuning, hyperfine, rabi
from bloqade import start, cast

# from bloqade.ir.location import Square, Chain
# import numpy as np
import pytest


def test_piecewise_const():
    prog = start.rydberg.detuning.uniform.piecewise_constant(
        durations=[0.1, 3.1, 0.05], values=[4, 4, 7.5]
    )

    ## inspect ir
    node1 = prog.__bloqade_ir__()
    ir1 = node1.waveforms[2]
    assert ir1.value == cast(7.5)
    assert ir1.duration == cast(0.05)

    ir2 = node1.waveforms[1]
    assert ir2.value == cast(4)
    assert ir2.duration == cast(3.1)

    ir3 = node1.waveforms[0]
    assert ir3.value == cast(4)
    assert ir3.duration == cast(0.1)


def test_registers():
    waveform = (
        ir.Linear("initial_detuning", "initial_detuning", "up_time")
        .append(ir.Linear("initial_detuning", "final_detuning", "anneal_time"))
        .append(ir.Linear("final_detuning", "final_detuning", "up_time"))
    )
    prog1 = start.rydberg.detuning.uniform.apply(waveform)
    reg = prog1.parse_register()

    assert reg.n_atoms == 0
    assert reg.n_dims is None


def test_scale():
    prog = start
    prog = (
        prog.rydberg.detuning.location(1)
        .scale(1.2)
        .piecewise_linear([0.1, 3.8, 0.1], [-10, -10, "a", "b"])
    )

    ## let Emit build ast
    seq = prog.parse_sequence()

    # print(type(list(seq.pulses.keys())[0]))
    Loc1 = list(seq.pulses[rydberg].fields[detuning].value.keys())[0]

    assert type(Loc1) == ir.ScaledLocations
    assert Loc1.value[ir.Location(1)] == cast(1.2)


def test_scale_location():
    prog = start.rydberg.detuning.location(1).scale(1.2).location(2).scale(3.3)

    assert prog._value == 3.3
    assert type(prog.__parent__) == spatial.Location
    assert prog.__parent__.__parent__._value == 1.2


def test_build_ast_Scale():
    prog = (
        start.rydberg.detuning.location(1)
        .scale(1.2)
        .location(2)
        .scale(3.3)
        .piecewise_constant(durations=[0.1], values=[1])
    )

    # compile ast:
    tmp = prog.parse_sequence()

    locs = list(tmp.pulses[rydberg].fields[detuning].value.keys())[0]
    wvfm = tmp.pulses[rydberg].fields[detuning].value[locs]

    assert locs == ir.ScaledLocations(
        {ir.Location(2): cast(3.3), ir.Location(1): cast(1.2)}
    )
    assert wvfm == ir.Constant(value=cast(1), duration=cast(0.1))


def test_spatial_var():
    prog = start.rydberg.detuning.var("a")

    assert prog._name == "a"

    prog = prog.piecewise_constant([0.1], [30])

    # test build ast:
    seq = prog.parse_sequence()

    assert seq.pulses[rydberg].fields[detuning].value[
        ir.RunTimeVector("a")
    ] == ir.Constant(value=30, duration=0.1)


def test_issue_107():
    waveform = (
        ir.Linear("initial_detuning", "initial_detuning", "up_time")
        .append(ir.Linear("initial_detuning", "final_detuning", "anneal_time"))
        .append(ir.Linear("final_detuning", "final_detuning", "up_time"))
    )

    prog1 = start.rydberg.detuning.uniform.apply(waveform)
    prog2 = start.rydberg.detuning.uniform.piecewise_linear(
        durations=["up_time", "anneal_time", "up_time"],
        values=[
            "initial_detuning",
            "initial_detuning",
            "final_detuning",
            "final_detuning",
        ],
    )

    assert prog1.parse_sequence() == prog2.parse_sequence()


def test_issue_150():
    prog = start.rydberg.detuning.uniform.linear(0, 1, 1).amplitude.uniform.linear(
        0, 2, 1
    )

    assert prog.parse_sequence() == ir.Sequence(
        {
            ir.rydberg: ir.Pulse(
                {
                    ir.detuning: ir.Field({ir.Uniform: ir.Linear(0, 1, 1)}),
                    ir.rabi.amplitude: ir.Field({ir.Uniform: ir.Linear(0, 2, 1)}),
                }
            )
        }
    )


def test_303_replicate_channel_should_add():
    prog = (
        start.rydberg.detuning.uniform.linear(0, 1, 1)
        .rabi.amplitude.uniform.linear(1, 2, 1)
        .detuning.uniform.linear(0, 2, 3)
    )

    assert prog.parse_sequence() == ir.Sequence(
        {
            ir.rydberg: ir.Pulse(
                {
                    ir.detuning: ir.Field(
                        {ir.Uniform: ir.Linear(0, 1, 1) + ir.Linear(0, 2, 3)}
                    ),
                    ir.rabi.amplitude: ir.Field({ir.Uniform: ir.Linear(1, 2, 1)}),
                }
            )
        }
    )

    prog1 = (
        start.rydberg.detuning.uniform.linear(0, 1, 1)
        .rabi.amplitude.uniform.linear(1, 2, 1)
        .rydberg.detuning.uniform.linear(0, 2, 3)
    )

    assert prog1.parse_sequence() == prog.parse_sequence()


def test_record():
    prog = start
    prog = (
        prog.rydberg.detuning.location(1)
        .piecewise_constant([0.1], [30])
        .record("detuning")
    )

    assert type(prog) == waveform.Record

    seq = prog.parse_sequence()
    assert seq.pulses[rydberg].fields[detuning].value[
        ir.ScaledLocations({ir.Location(1): cast(1)})
    ] == ir.Record(waveform=ir.Constant(value=30, duration=0.1), var=cast("detuning"))


def test_hyperfine_phase():
    prog = start.hyperfine.rabi.phase.location(1).piecewise_constant([0.1], [30])

    seq = prog.parse_sequence()

    assert seq.pulses[hyperfine].fields[rabi.phase].value[
        ir.ScaledLocations({ir.Location(1): cast(1)})
    ] == ir.Constant(value=30, duration=0.1)


def test_hyperfine_amplitude():
    prog = start.hyperfine.rabi.amplitude.location(1).piecewise_constant([0.1], [30])

    seq = prog.parse_sequence()

    assert seq.pulses[hyperfine].fields[rabi.amplitude].value[
        ir.ScaledLocations({ir.Location(1): cast(1)})
    ] == ir.Constant(value=30, duration=0.1)


def test_piecewise_constant_mismatch():
    with pytest.raises(AssertionError):
        start.hyperfine.rabi.amplitude.location(1).piecewise_constant([0.1, 0.5], [30])


def test_piecewise_linear_mismatch():
    with pytest.raises(AssertionError):
        start.hyperfine.rabi.amplitude.location(1).piecewise_linear(
            durations=[0.1, 0.5], values=[30, 20]
        )


def test_backend_route():
    prog = start.rydberg.detuning.uniform.constant(4, 4)

    assert isinstance(prog.device("quera.aquila"), quera.QuEraHardwareRoutine)
    assert isinstance(prog.device("braket.aquila"), braket.BraketHardwareRoutine)
    assert isinstance(
        prog.device("braket.local_emulator"), braket.BraketLocalEmulatorRoutine
    )
    with pytest.raises(ValueError):
        prog.device("foo")


def test_assign_error():
    import numpy as np

    with pytest.raises(ValueError):
        start.rydberg.detuning.uniform.constant("c", "t").assign(c=np, t=10)

    with pytest.raises(ValueError):
        start.rydberg.detuning.uniform.constant("c", "t").batch_assign(
            c=[1, 2, np], t=[10]
        )


def test_flatten_dupicate_error():
    with pytest.raises(ValueError):
        (
            start.add_position((0, 0))
            .rydberg.rabi.amplitude.uniform.constant("a", 1)
            .detuning.var("vec")
            .constant(1, 1)
            .assign(vec=[1])
            .flatten(["a", "a"])
            .braket.local_emulator()
            .run(1)
        )


def test_flatten_vector_error():
    with pytest.raises(ValueError):
        (
            start.add_position((0, 0))
            .rydberg.rabi.amplitude.uniform.constant(1, 1)
            .detuning.var("a")
            .constant(1, 1)
            .flatten(["a"])
            .braket.local_emulator()
            .run(1)
        )


def test_apply_sequence():
    sequence = start.rydberg.detuning.uniform.constant(1, 1).parse_sequence()

    start.add_position((0, 0)).apply(sequence).braket.local_emulator().run(1)


"""
prog = start
prog = (
    prog.rydberg.detuning.location(1)
    .location(2)
    .linear(start=1.0, stop=2.0, duration="x")
    .poly(coeffs=[1, 2, 3], duration="x")
    .location(3)
    .location(4)
    .constant(value=1.0, duration="x")
    .rabi.amplitude.location(5)
    .location(6)
    .linear(start=1.0, stop=2.0, duration="x")
    .phase.location(7)
    .constant(value=1.0, duration="x")
    .apply(ir.Linear(1.0, 2.0, "x"))
    .hyperfine.detuning.location(8)
    .poly(coeffs=[1, 2, 3], duration="x")
)
print(prog)

for idx in range(5):
    prog = prog.location(idx).linear(start=1.0, stop=2.0, duration="x")

print(prog)
print(prog.sequence)


job = (
    Square(4, lattice_spacing="a")
    .apply_defect_density(0.1)
    .rydberg.detuning.uniform.piecewise_linear(
        durations=[0.1, 3.8, 0.1], values=[-10, -10, "final_detuning", "final_detuning"]
    )
    .rabi.amplitude.uniform.piecewise_linear(
        durations=[0.1, 3.8, 0.1], values=[0.0, 15.0, 15.0, 0.0]
    )
    .assign(final_detuning=20, a=4)
    .mock(100)
)
print(job)

job = (
    Square(4, lattice_spacing="a")
    .apply_defect_count(4)
    .rydberg.detuning.uniform.piecewise_linear(
        durations=[0.1, 3.8, 0.1], values=[-10, -10, "final_detuning", "final_detuning"]
    )
    .rabi.amplitude.uniform.piecewise_linear(
        durations=[0.1, 3.8, 0.1], values=[0.0, 15.0, 15.0, 0.0]
    )
    .assign(final_detuning=20, a=4)
    .mock(100)
)
print(job)


job = (
    start.add_position((0, 0))
    .add_position((6, 0))
    .add_position((3, "distance"))
    .rydberg.detuning.uniform.piecewise_linear(
        durations=[0.1, 3.8, 0.1], values=[-10, -10, "final_detuning", "final_detuning"]
    )
    .rabi.amplitude.uniform.piecewise_linear(
        durations=[0.1, 3.8, 0.1], values=[0.0, 15.0, 15.0, 0.0]
    )
    .parallelize(20.0)
    .assign(final_detuning=20, distance=4)
    .mock(100)
)
print(job)

job = (
    start.add_position((0, 0))
    .add_position((6, 0))
    .add_position((3, "distance"))
    .rydberg.detuning.uniform.piecewise_linear(
        durations=[0.1, 3.8, 0.1], values=[-10, -10, "final_detuning", "final_detuning"]
    )
    .rabi.amplitude.uniform.piecewise_linear(
        durations=[0.1, 3.8, 0.1], values=[0.0, 15.0, 15.0, 0.0]
    )
    .parallelize(20.0)
    .assign(final_detuning=20, distance=4)
    .mock(100)
)

print(job)


def my_func(time, *, omega, phi=0, amplitude):
    return amplitude * np.cos(omega * time + phi)


durations = ir.cast([0.1, "run_time", 0.1])
total_duration = sum(durations)

job = (
    start.add_position((0, 0))
    .rydberg.detuning.uniform.fn(my_func, total_duration)
    .sample(0.05, "linear")
    .rydberg.rabi.amplitude.uniform.piecewise_linear(
        durations, [0, "rabi_max", "rabi_max", 0]
    )
    .assign(omega=15, amplitude=15, rabi_max=15)
    .batch_assign(run_time=np.linspace(0, 4.0, 101))
    .braket_local_simulator(1000)
)

print(job)


n_atoms = 11
atom_spacing = 6.1
run_time = var("run_time")
run_times = np.linspace(0.1, 4.0, 11)

quantum_scar_program = (
    Chain(11, lattice_spacing=atom_spacing)
    .rydberg.rabi.amplitude.uniform.piecewise_linear(
        [0.3, 1.6, 0.3], [0.0, 15.7, 15.7, 0.0]
    )
    .piecewise_linear([0.2, 1.4, 0.2], [0, 15.7, 15.7, 0])
    .slice(stop=run_time - 0.06)
    .record("rabi_value")
    .linear("rabi_value", 0, 0.06)
    .detuning.uniform.piecewise_linear([0.3, 1.6, 0.3], [-18.8, -18.8, 16.3, 16.3])
    .piecewise_linear([0.2, 1.6], [16.3, 0.0, 0.0])
    .slice(stop=run_time)
    .batch_assign(run_time=run_times)
    .braket_local_simulator(10000)
)


sequence = (
    start.rydberg.detuning.uniform.fn(my_func, total_duration)
    .sample(0.05, "linear")
    .rydberg.rabi.amplitude.uniform.piecewise_linear(
        durations, [0, "rabi_max", "rabi_max", 0]
    )
    .sequence
)

builder = start
for site in range(11):
    builder = builder.add_position((6 * site, 0))
    job = (
        builder.apply(sequence)
        .assign(omega=15, amplitude=15, rabi_max=15)
        .batch_assign(run_time=np.linspace(0, 4.0, 101))
        .braket_local_simulator(1000)
    )
    print(job)
"""



import bloqade.ir.location as location
import bloqade.ir.control.sequence as sequence
import bloqade.ir.control.pulse as pulse
import bloqade.ir.control.field as field
import bloqade.ir.control.waveform as waveform
import bloqade.ir.scalar as scalar
import bloqade.ir.analog_circuit as analog_circuit

from bloqade.ir.visitor.analog_circuit import AnalogCircuitVisitor
from bloqade.ir.visitor.location import LocationVisitor
from bloqade.ir.visitor.waveform import WaveformVisitor
from bloqade.ir.visitor.scalar import ScalarVisitor

import numbers
from typing import Any, Dict, Union
from decimal import Decimal


class AssignScalar(ScalarVisitor):
    def __init__(self, mapping: Dict[str, Decimal]):
        self.mapping = dict(mapping)

    def visit_literal(self, ast: scalar.Literal) -> scalar.Scalar:
        return ast

    def visit_variable(
        self, ast: scalar.Variable
    ) -> Union[scalar.Literal, scalar.Variable]:
        if ast.name in self.mapping:
            return scalar.AssignedVariable(ast.name, self.mapping[ast.name])

        return ast

    def visit_assigned_variable(
        self, ast: scalar.Variable
    ) -> Union[scalar.Literal, scalar.AssignedVariable]:
        if ast.name in self.mapping:
            raise ValueError(f"Variable {ast.name} already assigned to {ast.value}.")

        return ast

    def visit_negative(self, ast: scalar.Negative) -> scalar.Scalar:
        return -self.visit(ast.expr)

    def visit_add(self, ast: scalar.Add) -> scalar.Scalar:
        return self.visit(ast.lhs) + self.visit(ast.rhs)

    def visit_div(self, ast: scalar.Div) -> scalar.Scalar:
        return self.visit(ast.lhs) / self.visit(ast.rhs)

    def visit_mul(self, ast: scalar.Mul) -> scalar.Scalar:
        return self.visit(ast.lhs) * self.visit(ast.rhs)

    def visit_min(self, ast: scalar.Min) -> scalar.Scalar:
        return scalar.Scalar.canonicalize(
            scalar.Min(frozenset(list(map(self.visit, ast.scalars))))
        )

    def visit_max(self, ast: scalar.Min) -> scalar.Scalar:
        return scalar.Scalar.canonicalize(
            scalar.Max(frozenset(list(map(self.visit, ast.scalars))))
        )

    def emit(self, ast: scalar.Scalar) -> scalar.Scalar:
        # canonicalize at the end
        return scalar.Scalar.canonicalize(self.visit(ast))


class AssignWaveform(WaveformVisitor):
    def __init__(self, mapping: Dict[str, numbers.Real]):
        self.scalar_visitor = AssignScalar(mapping)
        self.mapping = dict(mapping)

    def visit_constant(self, ast: waveform.Constant) -> Any:
        value = self.scalar_visitor.emit(ast.value)
        duration = self.scalar_visitor.emit(ast.duration)
        return waveform.Constant(value, duration)

    def visit_linear(self, ast: waveform.Linear) -> Any:
        start = self.scalar_visitor.emit(ast.start)
        stop = self.scalar_visitor.emit(ast.stop)
        duration = self.scalar_visitor.emit(ast.duration)
        return waveform.Linear(start, stop, duration)

    def visit_poly(self, ast: waveform.Poly) -> Any:
        checkpoints = list(map(self.scalar_visitor.emit, ast.coeffs))
        duration = self.scalar_visitor.emit(ast.duration)
        return waveform.Poly(checkpoints, duration)

    def visit_python_fn(self, ast: waveform.PythonFn) -> Any:
        new_ast = waveform.PythonFn(ast.fn, self.scalar_visitor.emit(ast.duration))
        new_ast.parameters = list(map(self.scalar_visitor.emit, ast.parameters))
        return new_ast

    def visit_add(self, ast: waveform.Add) -> Any:
        return waveform.Add(self.visit(ast.left), self.visit(ast.right))

    def visit_alligned(self, ast: waveform.AlignedWaveform) -> Any:
        if isinstance(ast.value, scalar.Scalar):
            value = self.scalar_visitor.emit(ast.value)
        else:
            value = ast.value

        return waveform.AlignedWaveform(self.visit(ast.waveform), ast.alignment, value)

    def visit_append(self, ast: waveform.Append) -> Any:
        return waveform.Append(list(map(self.visit, ast.waveforms)))

    def visit_smooth(self, ast: waveform.Smooth) -> Any:
        static_radius = self.scalar_visitor.emit(ast.radius)
        return waveform.Smooth(static_radius, ast.kernel, self.visit(ast.waveform))

    def visit_negative(self, ast: waveform.Negative) -> Any:
        return waveform.Negative(self.visit(ast.waveform))

    def visit_record(self, ast: waveform.Record) -> Any:
        if ast.var.name in self.mapping:
            return self.visit(ast.waveform)
        else:
            return waveform.Record(self.visit(ast.waveform), ast.var)

    def visit_sample(self, ast: waveform.Sample) -> Any:
        dt = self.scalar_visitor.emit(ast.dt)
        return waveform.Sample(self.visit(ast.waveform), ast.interpolation, dt)

    def visit_scale(self, ast: waveform.Scale) -> Any:
        scale = self.scalar_visitor.emit(ast.scalar)
        return waveform.Scale(scale, self.visit(ast.waveform))

    def visit_slice(self, ast: waveform.Slice) -> Any:
        start = ast.interval.start
        stop = ast.interval.stop

        if start is not None:
            start = self.scalar_visitor.emit(start)

        if stop is not None:
            stop = self.scalar_visitor.emit(stop)

        return waveform.Slice(self.visit(ast.waveform), scalar.Interval(start, stop))

    def emit(self, ast: waveform.Waveform) -> waveform.Waveform:
        return waveform.Waveform.canonicalize(self.visit(ast))


class AssignLocation(LocationVisitor):
    def __init__(self, mapping: Dict[str, numbers.Real]):
        self.mapping = dict(mapping)
        self.scalar_visitor = AssignScalar(mapping)

    def visit_chain(self, ast: location.Chain) -> location.Chain:
        return location.Chain(
            ast.shape[0], self.scalar_visitor.emit(ast.lattice_spacing)
        )

    def visit_square(self, ast: location.Square) -> location.Square:
        return location.Square(
            ast.shape[0], self.scalar_visitor.emit(ast.lattice_spacing)
        )

    def visit_rectangular(self, ast: location.Rectangular) -> location.Rectangular:
        return location.Rectangular(
            ast.shape[0],
            ast.shape[1],
            self.scalar_visitor.emit(ast.lattice_spacing_x),
            self.scalar_visitor.emit(ast.lattice_spacing_y),
        )

    def visit_triangular(self, ast: location.Triangular) -> location.Triangular:
        return location.Triangular(
            ast.shape[0], self.scalar_visitor.emit(ast.lattice_spacing)
        )

    def visit_honeycomb(self, ast: location.Honeycomb) -> location.Honeycomb:
        return location.Honeycomb(
            ast.shape[0], self.scalar_visitor.emit(ast.lattice_spacing)
        )

    def visit_kagome(self, ast: location.Kagome) -> location.Kagome:
        return location.Kagome(
            ast.shape[0], self.scalar_visitor.emit(ast.lattice_spacing)
        )

    def visit_lieb(self, ast: location.Lieb) -> location.Lieb:
        return location.Lieb(
            ast.shape[0], self.scalar_visitor.emit(ast.lattice_spacing)
        )

    def visit_list_of_locations(
        self, ast: location.ListOfLocations
    ) -> location.ListOfLocations:
        return location.ListOfLocations(list(map(self.visit, ast.location_list)))

    def visit_location_info(self, ast: location.LocationInfo) -> location.LocationInfo:
        return location.LocationInfo(
            tuple(map(self.scalar_visitor.emit, ast.position)),
            bool(ast.filling.value),
        )

    def visit_parallel_register(
        self, ast: location.ParallelRegister
    ) -> location.ParallelRegister:
        return location.ParallelRegister(
            self.visit(ast._register), self.scalar_visitor.emit(ast._cluster_spacing)
        )


class AssignAnalogCircuit(AnalogCircuitVisitor):
    def __init__(self, mapping: Dict[str, numbers.Real]):
        self.waveform_visitor = AssignWaveform(mapping)
        self.scalar_visitor = AssignScalar(mapping)
        self.location_visitor = AssignLocation(mapping)
        self.mapping = dict(mapping)

    def visit_analog_circuit(self, ast: analog_circuit.AnalogCircuit) -> Any:
        return analog_circuit.AnalogCircuit(
            self.visit(ast.register), self.visit(ast.sequence)
        )

    def visit_register(self, ast: location.AtomArrangement) -> location.AtomArrangement:
        return self.location_visitor.visit(ast)

    def visit_parallel_register(
        self, ast: location.ParallelRegister
    ) -> location.ParallelRegister:
        return self.location_visitor.visit(ast)

    def visit_sequence(self, ast: sequence.SequenceExpr) -> sequence.SequenceExpr:
        match ast:
            case sequence.Sequence(pulses):
                return sequence.Sequence(
                    {
                        coupling_name: self.visit(sub_pulse)
                        for coupling_name, sub_pulse in pulses.items()
                    }
                )
            case sequence.Append(sequences):
                return sequence.Append(list(map(self.visit, sequences)))
            case sequence.Slice(sub_sequence, interval):
                return sequence.Slice(self.visit(sub_sequence), interval)
            case sequence.NamedSequence(sub_sequence, name):
                return sequence.NamedSequence(self.visit(sub_sequence), name)

    def visit_pulse(self, ast: pulse.PulseExpr) -> pulse.PulseExpr:
        match ast:
            case pulse.Pulse(fields):
                return pulse.Pulse(
                    {
                        field_name: self.visit(sub_field)
                        for field_name, sub_field in fields.items()
                    }
                )
            case pulse.Append(pulses):
                return pulse.Append(list(map(self.visit, pulses)))
            case pulse.Slice(sub_pulse, interval):
                return pulse.Slice(self.visit(sub_pulse), interval)
            case pulse.NamedPulse(name, sub_pulse):
                return pulse.NamedPulse(name, self.visit(sub_pulse))

    def visit_field(self, ast: field.Field) -> field.Field:
        return field.Field(
            {self.visit(sm): self.visit(wf) for sm, wf in ast.value.items()}
        )

    def visit_spatial_modulation(
        self, ast: field.SpatialModulation
    ) -> field.SpatialModulation:
        match ast:
            case field.UniformModulation():
                return ast
            case field.RunTimeVector(name):
                if name in self.mapping:
                    return field.AssignedRunTimeVector(name, self.mapping[name])
                else:
                    return ast

            case field.AssignedRunTimeVector(name, value):
                if name in self.mapping:
                    raise ValueError(f"Variable {name} already assigned to {value}.")
                return ast

            case field.ScaledLocations(values):
                return field.ScaledLocations(
                    {
                        loc: self.scalar_visitor.emit(scale)
                        for loc, scale in values.items()
                    }
                )

    def visit_waveform(self, ast: waveform.Waveform) -> waveform.Waveform:
        return self.waveform_visitor.emit(ast)

    def emit(self, ast: analog_circuit.AnalogCircuit) -> analog_circuit.AnalogCircuit:
        return self.visit(ast)



from ..tree_print import Printer
from ..scalar import Scalar, Interval, Literal, Variable, AssignedVariable, cast, var

from bisect import bisect_left
from dataclasses import InitVar
from decimal import Decimal
from pydantic.dataclasses import dataclass
from typing import Any, Tuple, Union, List, Callable
from enum import Enum

from bokeh.plotting import figure, show
import numpy as np
import inspect
import scipy.integrate as integrate


def instruction(duration: Any) -> "PythonFn":
    # turn python function into a waveform instruction."""

    def waveform_wrapper(fn: Callable) -> "PythonFn":
        return PythonFn(fn, duration)

    return waveform_wrapper


class AlignedValue(str, Enum):
    Left = "left_value"
    Right = "right_value"


class Alignment(str, Enum):
    Left = "left_aligned"
    Right = "right_aligned"


@dataclass
class Waveform:
    """
    Waveform node in the IR.

    - [`<instruction>`][bloqade.ir.control.waveform.Instruction]
    - [`<smooth>`][bloqade.ir.control.waveform.Smooth]
    - [`<slice>`][bloqade.ir.control.waveform.Slice]
    - [`<apppend>`][bloqade.ir.control.waveform.Append]
    - [`<negative>`][bloqade.ir.control.waveform.Negative]
    - [`<scale>`][bloqade.ir.control.waveform.Scale]
    - [`<add>`][bloqade.ir.control.waveform.Add]
    - [`<record>`][bloqade.ir.control.waveform.Record]
    - [`<sample>`][bloqade.ir.control.waveform.Sample]

    ```bnf
    <waveform> ::= <instruction>
        | <smooth>
        | <slice>
        | <append>
        | <negative>
        | <scale>
        | <add>
        | <record>
        | <sample>
    ```
    """

    def __post_init__(self):
        self._duration = None

    def __call__(self, clock_s: float, **kwargs) -> float:
        return float(self.eval_decimal(Decimal(str(clock_s)), **kwargs))

    def __mul__(self, scalar: Scalar) -> "Waveform":
        return self.scale(cast(scalar))

    def __rmul__(self, scalar: Scalar) -> "Waveform":
        return self.scale(cast(scalar))

    def __add__(self, other: "Waveform"):
        return self.add(other)

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        raise NotImplementedError

    def add(self, other: "Waveform") -> "Waveform":
        return self.canonicalize(Add(self, other))

    def append(self, other: "Waveform") -> "Waveform":
        return self.canonicalize(Append([self, other]))

    def figure(self, **assignments):
        """get figure of the plotting the waveform.

        Returns:
            figure: a bokeh figure
        """
        # Varlist = []
        duration = float(self.duration(**assignments))
        times = np.linspace(0, duration, 1001)
        values = [self.__call__(time, **assignments) for time in times]
        fig = figure(
            sizing_mode="stretch_both",
            x_axis_label="Time (s)",
            y_axis_label="Waveform(t)",
            tools="hover",
        )
        fig.line(times, values)

        return fig

    def _get_data(self, npoints, **assignments):
        duration = float(self.duration(**assignments))
        times = np.linspace(0, duration, npoints + 1)
        values = [self.__call__(time, **assignments) for time in times]
        return times, values

    def show(self, **assignments):
        show(self.figure(**assignments))

    def align(
        self, alignment: Alignment, value: Union[None, AlignedValue, Scalar] = None
    ) -> "Waveform":
        if value is None:
            if alignment == Alignment.Left:
                value = AlignedValue.Left
            elif alignment == Alignment.Right:
                value = AlignedValue.Right
            else:
                raise ValueError(f"Invalid alignment: {alignment}")
        elif value is Scalar or value is AlignedValue:
            return self.canonicalize(AlignedWaveform(self, alignment, value))
        else:
            return self.canonicalize(AlignedWaveform(self, alignment, cast(value)))

    def smooth(self, radius, kernel: "SmoothingKernel") -> "Waveform":
        return self.canonicalize(
            Smooth(kernel=kernel, waveform=self, radius=cast(radius))
        )

    def scale(self, value) -> "Waveform":
        return self.canonicalize(Scale(cast(value), self))

    def __neg__(self) -> "Waveform":
        return self.canonicalize(Negative(self))

    def __getitem__(self, s: slice) -> "Waveform":
        return self.canonicalize(Slice(self, Interval.from_slice(s)))

    def record(self, variable_name: Union[str, Variable]):
        return Record(self, cast(variable_name))

    @property
    def duration(self) -> Scalar:
        if self._duration:
            return self._duration

        match self:
            case Instruction(duration=duration):
                self._duration = duration
            case AlignedWaveform(waveform=waveform, alignment=_, value=_):
                self._duration = waveform.duration()
            case Slice(waveform=waveform, interval=interval):
                match (interval.start, interval.stop):
                    case (None, None):
                        raise ValueError(f"Cannot compute duration of {self}")
                    case (start, None):
                        self._duration = waveform.duration - start
                    case (None, stop):
                        self._duration = stop
                    case (start, stop):
                        self._duration = stop - start
            case Append(waveforms=waveforms):
                duration = cast(0.0)
                for waveform in waveforms:
                    duration = duration + cast(waveform.duration)

                self._duration = duration
            case Sample(waveform=waveform, interpolation=_, dt=_):
                self._duration = waveform.duration
            case Smooth(waveform=waveform, kernel=_, radius=_):
                return waveform.duration
            case Record(waveform=waveform, var=_):
                return waveform.duration
            case _:
                raise ValueError(f"Cannot compute duration of {self}")
        return self._duration

    @staticmethod
    def canonicalize(expr: "Waveform") -> "Waveform":
        match expr:
            case Append([Append(lhs), Append(rhs)]):
                return Append(list(map(Waveform.canonicalize, lhs + rhs)))
            case Append([Append(waveforms), waveform]):
                return Waveform.canonicalize(Append(waveforms + [waveform]))
            case Append([waveform, Append(waveforms)]):
                return Waveform.canonicalize(Append([waveform] + waveforms))
            case _:
                return expr

    def __repr__(self) -> str:
        ph = Printer()
        ph.print(self)
        return ph.get_value()

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    def print_node(self):
        raise NotImplementedError


@dataclass(repr=False)
class AlignedWaveform(Waveform):
    """

    ```bnf
    <padded waveform> ::= <waveform> | <waveform> <alignment> <value>

    <alignment> ::= 'left aligned' | 'right aligned'
    <value> ::= 'left value' | 'right value' | <scalar expr>
    ```
    """

    waveform: Waveform
    alignment: Alignment
    value: Union[Scalar, AlignedValue]

    def print_node(self):
        return "AlignedWaveform"

    def children(self):
        annotated_children = {}
        annotated_children["Waveform"] = self.waveform

        match self.alignment:
            case Alignment.Left:
                annotated_children["Alignment"] = "Left"
            case Alignment.Right:
                annotated_children["Alignment"] = "Right"

        match self.value:
            case Scalar():
                annotated_children["Value"] = self.value
            case AlignedValue.Left:
                annotated_children["Value"] = "Left"
            case AlignedValue.Right:
                annotated_children["Value"] = "Right"

        return annotated_children


@dataclass(repr=False)
class Instruction(Waveform):
    """Instruction node in the IR.

    - [`<linear>`][bloqade.ir.control.waveform.Linear]
    - [`<constant>`][bloqade.ir.control.waveform.Constant]
    - [`<poly>`][bloqade.ir.control.waveform.Poly]
    - [`<python-fn>`][bloqade.ir.control.waveform.PythonFn]


    ```bnf
    <instruction> ::= <linear>
        | <constant>
        | <poly>
        | <python-fn>
    ```
    """

    pass


@dataclass(init=False, repr=False)
class Linear(Instruction):
    """
    ```bnf
    <linear> ::= 'linear' <scalar expr> <scalar expr>
    ```

    f(t=0:duration) = start + (stop-start)/duration * t

    Args:
        start (Scalar): start value
        stop (Scalar): stop value
        duration (Scalar): the time span of the linear waveform.

    """

    start: Scalar
    stop: Scalar
    duration: InitVar[Scalar]

    def __init__(self, start, stop, duration):
        self.start = cast(start)
        self.stop = cast(stop)
        self._duration = cast(duration)

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        start_value = self.start(**kwargs)
        stop_value = self.stop(**kwargs)

        if clock_s > self.duration(**kwargs):
            return Decimal(0)
        else:
            return (
                (stop_value - start_value) / self.duration(**kwargs)
            ) * clock_s + start_value

    def __str__(self):
        return (
            f"Linear(start={str(self.start)}, stop={str(self.stop)}, "
            f"duration={str(self.duration)})"
        )

    def print_node(self):
        return "Linear"

    def children(self):
        return {"start": self.start, "stop": self.stop, "duration": self.duration}


@dataclass(init=False, repr=False)
class Constant(Instruction):
    """
    ```bnf
    <constant> ::= 'constant' <scalar expr>
    ```

    f(t=0:duration) = value

    Args:
        value (Scalar): the constant value
        duration (Scalar): the time span of the constant waveform.

    """

    value: Scalar
    duration: InitVar[Scalar]

    def __init__(self, value, duration):
        self.value = cast(value)
        self._duration = cast(duration)

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        constant_value = self.value(**kwargs)
        if clock_s > self.duration(**kwargs):
            return Decimal(0)
        else:
            return constant_value

    def __str__(self):
        return f"Constant(value={str(self.value)}, duration={str(self.duration)})"

    def print_node(self):
        return "Constant"

    def children(self):
        return {"value": self.value, "duration": self.duration}


@dataclass(init=False, repr=False)
class Poly(Instruction):
    """
    ```bnf
    <poly> ::= <scalar>+
    ```

    f(t=0:duration) = c[0] + c[1]t + c[2]t^2 + ... + c[n-1]t^n-1 + c[n]t^n

    Args:
        checkpoints (List[Scalar]): the coefficients c[] of the polynomial.
        duration (Scalar): the time span of the waveform.

    """

    coeffs: List[Scalar]
    duration: InitVar[Scalar]

    def __init__(self, coeffs, duration):
        self.coeffs = list(map(cast, coeffs))
        self._duration = cast(duration)

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        # b + x + x^2 + ... + x^n-1 + x^n
        if clock_s > self.duration(**kwargs):
            return Decimal(0)
        else:
            # call clock_s on each element of the scalars,
            # then apply the proper powers
            value = Decimal(0)
            for exponent, scalar_expr in enumerate(self.coeffs):
                value += scalar_expr(**kwargs) * clock_s**exponent

            return value

    def __str__(self):
        return f"Poly({str(self.coeffs)}, {str(self.duration)})"

    def print_node(self) -> str:
        return "Poly"

    def children(self):
        # should have annotation for duration
        # then annotations for the polynomial terms and exponents

        annotated_checkpoints = {}
        for i, checkpoint in enumerate(self.coeffs):
            if i == 0:
                annotated_checkpoints["b"] = checkpoint
            elif i == 1:
                annotated_checkpoints["t"] = checkpoint
            else:
                annotated_checkpoints["t^" + str(i)] = checkpoint

        annotated_checkpoints["duration"] = self._duration

        return annotated_checkpoints


@dataclass(init=False, repr=False)
class PythonFn(Instruction):
    """

    ```bnf
    <python-fn> ::= 'python-fn' <python function def> <scalar expr>
    ```
    """

    fn: Callable  # [[float, ...], float] # f(t) -> value
    parameters: List[
        Union[Variable, AssignedVariable, Literal]
    ]  # come from ast inspect
    duration: InitVar[Scalar]

    def __init__(self, fn: Callable, duration: Any):
        self.fn = fn
        self._duration = cast(duration)

        signature = inspect.getfullargspec(fn)

        if signature.varargs is not None:
            raise ValueError("Cannot have varargs")

        if signature.varkw is not None:
            raise ValueError("Cannot have varkw")

        # get default kwonly first:
        default_variables = []
        if signature.kwonlydefaults is not None:
            default_variables = [
                AssignedVariable(name, cast(Decimal(str(value))))
                for name, value in signature.kwonlydefaults.items()
            ]

        variables = signature.args[1:] + [
            v for v in signature.kwonlyargs if v not in signature.kwonlydefaults.keys()
        ]
        self.parameters = list(map(var, variables)) + default_variables

    def eval_decimal(self, clock_s: Decimal, **assignments) -> Decimal:
        if clock_s > self.duration(**assignments):
            return Decimal(0)

        kwargs = {param.name: float(param(**assignments)) for param in self.parameters}

        return Decimal(
            str(
                self.fn(
                    float(clock_s),
                    **kwargs,
                )
            )
        )

    def print_node(self):
        return f"PythonFn: {self.fn.__name__}"

    def children(self):
        return {"duration": self.duration}


@dataclass(init=False)
class SmoothingKernel:
    def __call__(self, value: float) -> float:
        raise NotImplementedError


class FiniteSmoothingKernel(SmoothingKernel):
    # kernel that is zero outside of (-1, 1)
    pass


class InfiniteSmoothingKernel(SmoothingKernel):
    # Kernel that is non-zero for all values
    pass


class Gaussian(InfiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return np.exp(-(value**2) / 2) / np.sqrt(2 * np.pi)


class Logistic(InfiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return np.exp(-(np.logaddexp(0, value) + np.logaddexp(0, -value)))


class Sigmoid(InfiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return (2 / np.pi) * np.exp(-np.logaddexp(-value, value))


class Triangle(FiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return np.maximum(0, 1 - np.abs(value))


class Uniform(FiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return np.asarray(np.abs(value) <= 1, dtype=np.float64).squeeze()


class Parabolic(FiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return (3 / 4) * np.maximum(0, 1 - value**2)


class Biweight(FiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return (15 / 16) * np.maximum(0, 1 - value**2) ** 2


class Triweight(FiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return (35 / 32) * np.maximum(0, 1 - value**2) ** 3


class Tricube(FiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return (70 / 81) * np.maximum(0, 1 - np.abs(value) ** 3) ** 3


class Cosine(FiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return np.maximum(0, np.pi / 4 * np.cos(np.pi / 2 * value))


GaussianKernel = Gaussian()
LogisticKernel = Logistic()
SigmoidKernel = Sigmoid()
TriangleKernel = Triangle()
UniformKernel = Uniform()
ParabolicKernel = Parabolic()
BiweightKernel = Biweight()
TriweightKernel = Triweight()
TricubeKernel = Tricube()
CosineKernel = Cosine()


@dataclass(repr=False)
class Smooth(Waveform):
    """
    ```bnf
    <smooth> ::= 'smooth' <kernel> <waveform>
    ```
    """

    radius: Scalar
    kernel: SmoothingKernel
    waveform: Waveform

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        float_clock_s = float(clock_s)
        radius = float(self.radius(**kwargs))
        duration = float(self.duration(**kwargs))
        waveform_start = self.waveform(0, **kwargs)
        waveform_stop = self.waveform(float(duration), **kwargs)

        def waveform(clock: float) -> float:
            if clock < 0:
                return waveform_start
            elif clock > duration:
                return waveform_stop
            else:
                return self.waveform(clock, **kwargs)

        def integrade(s):
            return self.kernel(s) * waveform(radius * s + float_clock_s)

        if isinstance(self.kernel, FiniteSmoothingKernel):
            return integrate.quad(integrade, -1, 1, epsabs=1e-4, epsrel=1e-4)[0]
        elif isinstance(self.kernel, InfiniteSmoothingKernel):
            return integrate.quad(integrade, -np.inf, np.inf, epsabs=1e-4, epsrel=1e-4)[
                0
            ]
        else:
            raise ValueError(f"Invalid kernel: {self.kernel}")

    def __str__(self):
        return f"Smooth(kernel={str(self.kernel)}, waveform={str(self.waveform)})"


@dataclass(repr=False)
class Slice(Waveform):
    """
    ```
    <slice> ::= <waveform> <scalar.interval>
    ```
    """

    waveform: Waveform
    interval: Interval

    def __str__(self):
        return f"{str(self.waveform)}[{str(self.interval)}]"

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        if clock_s > self.duration(**kwargs):
            return Decimal(0)
        if self.interval.start is None:
            start_time = Decimal(0)
        else:
            start_time = self.interval.start(**kwargs)
        return self.waveform.eval_decimal(clock_s + start_time, **kwargs)

    def print_node(self):
        return "Slice"

    def children(self):
        return [self.waveform, self.interval]


@dataclass(repr=False)
class Append(Waveform):
    """
    ```bnf
    <append> ::= <waveform>+
    ```
    """

    waveforms: List[Waveform]

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        append_time = Decimal(0)
        for waveform in self.waveforms:
            duration = waveform.duration(**kwargs)

            if clock_s <= append_time + duration:
                return waveform.eval_decimal(clock_s - append_time, **kwargs)

            append_time += duration

        return Decimal(0)

    def __str__(self):
        return f"waveform.Append(waveforms={str(self.waveforms)})"

    def print_node(self):
        return "Append"

    def children(self):
        return self.waveforms


@dataclass(repr=False)
class Negative(Waveform):
    """
    ```bnf
    <negative> ::= '-' <waveform>
    ```
    """

    waveform: Waveform

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        return -self.waveform.eval_decimal(clock_s, **kwargs)

    def __str__(self):
        return f"-({str(self.waveform)})"

    def print_node(self):
        return "Negative"

    def children(self):
        return [self.waveform]


@dataclass(init=False, repr=False)
class Scale(Waveform):
    """
    ```bnf
    <scale> ::= <scalar expr> '*' <waveform>
    ```
    """

    scalar: Scalar
    waveform: Waveform

    def __init__(self, scalar, waveform: Waveform):
        self.scalar = cast(scalar)
        self.waveform = waveform

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        return self.scalar(**kwargs) * self.waveform.eval_decimal(clock_s, **kwargs)

    def __str__(self):
        return f"({str(self.scalar)} * {str(self.waveform)})"

    def print_node(self):
        return "Scale"

    def children(self):
        return [self.scalar, self.waveform]


@dataclass(repr=False)
class Add(Waveform):
    """
    ```bnf
    <add> ::= <waveform> '+' <waveform>
    ```
    """

    left: Waveform
    right: Waveform

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        return self.left(clock_s, **kwargs) + self.right(clock_s, **kwargs)

    def __str__(self):
        return f"({str(self.left)} + {str(self.right)})"

    def print_node(self):
        return "+"

    def children(self):
        return [self.left, self.right]


@dataclass(repr=False)
class Record(Waveform):
    """
    ```bnf
    <record> ::= 'record' <waveform> <var>
    ```
    """

    waveform: Waveform
    var: Variable

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        return self.waveform(clock_s, **kwargs)

    def print_node(self):
        return "Record"

    def children(self):
        return {"Waveform": self.waveform, "Variable": self.var}

    def __str__(self):
        return f"Record({str(self.waveform)}, {str(self.var)})"


class Interpolation(str, Enum):
    Linear = "linear"
    Constant = "constant"


@dataclass(repr=False)
class Sample(Waveform):
    """
    ```bnf
    <sample> ::= 'sample' <waveform> <interpolation> <scalar>
    ```
    """

    waveform: Waveform
    interpolation: Interpolation
    dt: Scalar

    def samples(self, **kwargs) -> Tuple[List[Decimal], List[Decimal]]:
        duration = self.duration(**kwargs)
        dt = self.dt(**kwargs)

        clock = Decimal("0.0")
        clocks = []
        values = []
        while clock <= duration - dt:
            values.append(self.waveform.eval_decimal(clock, **kwargs))
            clocks.append(clock)
            clock += dt

        values.append(self.waveform.eval_decimal(duration, **kwargs))
        clocks.append(duration)

        return clocks, values

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        times, values = self.samples(**kwargs)
        i = bisect_left(times, clock_s)

        if i == len(times):
            return Decimal(0)

        match self.interpolation:
            case Interpolation.Linear:
                if i == 0:
                    return values[i]
                else:
                    slope = (values[i] - values[i - 1]) / (times[i] - times[i - 1])
                    return slope * (clock_s - times[i - 1]) + values[i - 1]

            case Interpolation.Constant:
                if i == 0:
                    return values[i]
                else:
                    return values[i - 1]
            case _:
                raise ValueError("No interpolation specified")

    def print_node(self):
        return f"Sample {self.interpolation.value}"

    def children(self):
        return {"Waveform": self.waveform, "sample_step": self.dt}



from bloqade.task.quera import QuEraTask
from bloqade.task.braket import BraketTask
from bloqade.submission.quera import QuEraBackend
from bloqade.submission.braket import BraketBackend
import pytest


def test_quera_task():
    backend = QuEraBackend(api_hostname="a", qpu_id="b")
    task = QuEraTask(
        task_ir=None, task_id=None, metadata={}, backend=backend, parallel_decoder=None
    )

    with pytest.raises(ValueError):
        task.fetch()

    assert task._result_exists() is False

    with pytest.raises(ValueError):
        task.pull()


def test_braket_task():
    backend = BraketBackend()
    task = BraketTask(
        task_ir=None, task_id=None, metadata={}, backend=backend, parallel_decoder=None
    )

    with pytest.raises(ValueError):
        task.fetch()

    assert task._result_exists() is False

    with pytest.raises(ValueError):
        task.pull()


#
# batch = RemoteBatch.from_json(...)
# batch.fetch() # update results,
#               # this is non-blocking.
#               # It only pull results if the remote job complete

# batch.pull() # this is blocking. it will hanging there waiting for job to complete.
# batch.to_json(...)

# # Get finished tasks (not nessasry sucessfully completed)
# finished_batch = batch.get_finished_tasks()

# # Get complete tasks (sucessfully completed)
# completed_batch = batch.get_completed_tasks()

# # Remove failed tasks:
# new_batch = batch.remove_failed_tasks()



from bloqade import cast, var
import bloqade.ir.scalar as scalar
import pytest
from decimal import Decimal
from io import StringIO
from IPython.lib.pretty import PrettyPrinter as PP
import bloqade.ir.tree_print as trp

trp.color_enabled = False


def test_var():
    assert var("a") == scalar.Variable("a")
    assert var("a") != scalar.Variable("b")

    with pytest.raises(TypeError):
        var(1)

    with pytest.raises(ValueError):
        var("a*b")

    Vv = var("a")
    vs = var(Vv)

    assert vs == Vv

    mystdout = StringIO()
    p = PP(mystdout)
    Vv._repr_pretty_(p, 0)

    assert mystdout.getvalue() == "Variable: a\n"


def test_cast():
    assert cast("a") == scalar.Variable("a")
    assert cast(1) == scalar.Literal(Decimal("1"))
    assert cast(1.20391023) == scalar.Literal(Decimal("1.20391023"))

    with pytest.raises(ValueError):
        cast("a-b")

    mystdout = StringIO()
    p = PP(mystdout)
    cast(1)._repr_pretty_(p, 0)

    assert mystdout.getvalue() == "Literal: 1\n"


def test_add():
    assert var("a") + var("b") == scalar.Add(var("a"), var("b"))
    assert var("a") + 1 == scalar.Add(var("a"), scalar.Literal(Decimal("1")))
    assert 0 + var("a") == var("a")
    assert var("a") + 0 == var("a")
    assert 1 + var("a") == scalar.Add(scalar.Literal(Decimal("1")), var("a"))
    assert cast(1) + cast(2) == scalar.Literal(Decimal("3"))

    mystdout = StringIO()
    p = PP(mystdout)
    (1 + var("a"))._repr_pretty_(p, 0)

    assert mystdout.getvalue() == "+\n├─ Literal: 1\n⋮\n└─ Variable: a⋮\n"


def test_sub():
    assert var("a") - var("b") == scalar.Add(var("a"), scalar.Negative(var("b")))
    assert var("a") - 1 == scalar.Add(var("a"), scalar.Negative(cast(1)))
    assert 1 - var("a") == scalar.Add(
        scalar.Literal(Decimal("1")), scalar.Negative(var("a"))
    )
    assert cast(1) - cast(2) == scalar.Literal(Decimal("-1"))
    assert -cast(1) + cast(2) == scalar.Literal(Decimal("1"))


def test_mul():
    assert var("a") * var("b") == scalar.Mul(var("a"), var("b"))
    assert 1 * var("a") == var("a")
    assert var("a") * 1 == var("a")
    assert 3 * var("a") == scalar.Mul(scalar.Literal(Decimal("3")), var("a"))
    assert cast(1) * cast(2) == scalar.Literal(Decimal("2"))

    mystdout = StringIO()
    p = PP(mystdout)
    (3 * var("a"))._repr_pretty_(p, 0)

    assert mystdout.getvalue() == "*\n├─ Literal: 3\n⋮\n└─ Variable: a⋮\n"


def test_div():
    assert var("a") / var("b") == scalar.Div(var("a"), var("b"))
    assert 1 / var("a") == scalar.Div(cast(1), var("a"))
    assert var("a") / 1 == var("a")
    assert 3 / var("a") == scalar.Div(cast(3), var("a"))
    assert cast(1) / cast(2) == cast(0.5)

    mystdout = StringIO()
    p = PP(mystdout)
    (3 / var("a"))._repr_pretty_(p, 0)

    assert mystdout.getvalue() == "/\n├─ Literal: 3\n⋮\n└─ Variable: a⋮\n"


@pytest.mark.skip(reason="no longer supported")
def test_list_of_var():
    pylist = ["a", "b", "c"]
    vlist = var(pylist)
    for pyobj, bobj in zip(pylist, vlist):
        assert bobj == var(pyobj)


@pytest.mark.skip(reason="no longer supported")
def test_tuple_of_var():
    pylist = ("a", "b", "c")
    vlist = var(pylist)
    for pyobj, bobj in zip(pylist, vlist):
        assert bobj == var(pyobj)


def test_var_member():
    va = var("a")

    assert va.children() == []
    assert va.print_node() == "Variable: a"


def test_scalar_var():
    va = cast(1.0)
    with pytest.raises(TypeError):
        var(va)


@pytest.mark.skip(reason="no longer supported")
def test_invalid_keyword():
    with pytest.raises(ValueError):
        var("config_file")

    with pytest.raises(ValueError):
        var("clock_s")


def test_negative_node():
    sa = cast(1.0)
    nsa = scalar.Negative(sa)

    assert nsa.children() == [sa]
    assert nsa.print_node() == "Negative"

    assert str(nsa) == "-(1.0)"


# def test_base_invalid():


def test_interval():
    itvl = scalar.Interval(start=cast(0.0), stop=cast(1.0))
    itvl_no_start = scalar.Interval(None, stop=cast(1.0))
    itvl_no_stop = scalar.Interval(start=cast(0), stop=None)
    itvl_wrong = scalar.Interval(None, None)

    assert itvl.print_node() == "Interval"

    with pytest.raises(ValueError):
        str(itvl_wrong)

    assert str(itvl) == "0.0:1.0"
    assert str(itvl_no_start) == ":1.0"
    assert str(itvl_no_stop) == "0:"

    with pytest.raises(ValueError):
        itvl_wrong.children()

    assert itvl.children() == {"start": cast(0.0), "stop": cast(1.0)}
    assert itvl_no_start.children() == {"stop": cast(1.0)}
    assert itvl_no_stop.children() == {"start": cast(0)}


def test_canonicalize_neg_neg():
    sa = cast(1.0)
    nsa = -sa
    nsa2 = -nsa

    out2 = scalar.Scalar.canonicalize(nsa2)
    assert out2.value == Decimal("1.0")

    sb = cast(-1.0)
    nsb = scalar.Negative(sb)

    out3 = scalar.Scalar.canonicalize(nsb)
    assert out3 == cast(1.0)


def test_canonicalize_add_neg():
    sa = cast(1.0)
    nsa = scalar.Negative(sa)

    outR = scalar.Add(sa, nsa)
    outL = scalar.Add(nsa, sa)

    outR = scalar.Scalar.canonicalize(outR)
    outL = scalar.Scalar.canonicalize(outL)

    assert outR == scalar.Literal(0.0)
    assert outL == scalar.Literal(0.0)

    A = cast(1.0)
    B = cast(2.0)

    C = scalar.Add(A, B)
    nC = -C

    assert scalar.Scalar.canonicalize(scalar.Add(C, nC)) == scalar.Literal(0.0)
    assert scalar.Scalar.canonicalize(scalar.Add(nC, C)) == scalar.Literal(0.0)


def test_canonicalize_mul_zero():
    zero = scalar.Literal(0.0)
    one = cast(1)

    outR = scalar.Mul(zero, one)
    outL = scalar.Mul(one, zero)

    outR = scalar.Scalar.canonicalize(outR)
    outL = scalar.Scalar.canonicalize(outL)

    assert outR == scalar.Literal(0.0)
    assert outL == scalar.Literal(0.0)

    A = cast(1.0)
    B = cast(2.0)

    C = scalar.Add(A, B)

    assert scalar.Scalar.canonicalize(scalar.Mul(zero, C)) == scalar.Literal(0.0)
    assert scalar.Scalar.canonicalize(scalar.Mul(C, zero)) == scalar.Literal(0.0)


def test_add_scalar():
    A = cast(1)
    B = cast(2)

    C = scalar.Add(A, B)

    assert C.children() == [A, B]
    assert C.print_node() == "+"
    assert str(C) == "(1 + 2)"


def test_add_zero():
    A = cast(1)
    B = cast(2)

    C = scalar.Add(A, B)  # expression
    zero = cast(0.0)

    LC = scalar.Scalar.canonicalize(scalar.Add(C, zero))
    RC = scalar.Scalar.canonicalize(scalar.Add(zero, C))

    assert LC == cast(3)
    assert RC == cast(3)


def test_mul_one():
    A = cast(2)
    B = cast(2)

    C = scalar.Add(A, B)  # expression
    one = cast(1.0)

    LC = scalar.Scalar.canonicalize(scalar.Mul(C, one))
    RC = scalar.Scalar.canonicalize(scalar.Mul(one, C))

    assert LC == cast(4)
    assert RC == cast(4)


def test_div_rone():
    A = cast(2)
    B = cast(2)

    C = scalar.Add(A, B)  # expression
    one = cast(1.0)

    LC = scalar.Scalar.canonicalize(scalar.Div(C, one))

    assert LC == cast(4)

    dC = A / B
    assert dC == cast(1)


def test_mul_scalar():
    A = cast(1)
    B = cast(2)

    C = scalar.Mul(A, B)

    assert C.children() == [A, B]
    assert C.print_node() == "*"
    assert str(C) == "(1 * 2)"


def test_div_scalar():
    A = cast(1)
    B = cast(2)

    C = scalar.Div(A, B)

    assert C.children() == [A, B]
    assert C.print_node() == "/"
    assert str(C) == "(1 / 2)"

    assert C() == 0.5


def test_min_scalar():
    A = cast(1)
    B = cast(2)
    C = cast(3)

    D = scalar.Min([A, B, C])
    assert D.children() == [A, B, C]
    assert D.print_node() == "min"
    assert str(D) == "scalar.Min(frozenset({1, 2, 3}))"

    mystdout = StringIO()
    p = PP(mystdout)
    D._repr_pretty_(p, 0)

    assert (
        mystdout.getvalue()
        == "min\n├─ Literal: 1\n⋮\n├─ Literal: 2\n⋮\n└─ Literal: 3⋮\n"
    )


def test_max_scalar():
    A = cast(1)
    B = cast(2)
    C = cast(3)

    D = scalar.Max([A, B, C])
    assert D.children() == [A, B, C]
    assert D.print_node() == "max"
    assert str(D) == "scalar.Max(frozenset({1, 2, 3}))"

    mystdout = StringIO()
    p = PP(mystdout)
    D._repr_pretty_(p, 0)

    assert (
        mystdout.getvalue()
        == "max\n├─ Literal: 1\n⋮\n├─ Literal: 2\n⋮\n└─ Literal: 3⋮\n"
    )


def test_cast_decimal():
    A = cast(Decimal("1.0"))

    assert A.value == Decimal("1.0")


def test_Interval_from_pyslice():
    with pytest.raises(ValueError):
        scalar.Interval.from_slice(slice(None, None, None))

    with pytest.raises(ValueError):
        scalar.Interval.from_slice(slice(None, 3, 3))

    with pytest.raises(ValueError):
        scalar.Interval.from_slice(slice(3, None, 3))

    with pytest.raises(ValueError):
        scalar.Interval.from_slice(slice(3, 3, 3))

    with pytest.raises(ValueError):
        scalar.Interval.from_slice(slice(None, None, 3))

    itvl = scalar.Interval.from_slice(slice(5, 6, None))
    assert itvl.start == cast(5)
    assert itvl.stop == cast(6)


def test_Slice():
    itvl = scalar.Interval(cast(5), cast(6))

    slc = scalar.Slice(cast(1), itvl)

    assert str(slc) == "1[5:6]"
    assert slc.children() == {"Scalar": cast(1), None: itvl}
    assert slc.print_node() == "Slice"

    mystdout = StringIO()
    p = PP(mystdout)

    slc._repr_pretty_(p, 0)

    assert (
        mystdout.getvalue()
        == "Slice\n"
        + "├─ Scalar\n"
        + "│  ⇒ Literal: 1\n"
        + "⋮\n"
        + "└─ Interval\n"
        + "⋮\n"
    )


def test_assigned_variable_methods():
    assigned_var = scalar.AssignedVariable("a", Decimal("1.0"))

    assert assigned_var.children() == []
    assert assigned_var.print_node() == "AssignedVariable: a = 1.0"
    assert str(assigned_var) == "a"


@pytest.mark.skip(reason="Changed implementation of static_assign")
def test_static_assign():
    literal = cast(1.0)
    assert literal == literal.static_assign(a=1, b=2)

    float_value = 1.394023

    variable = var("A")
    assert variable == variable.static_assign(b=2)
    assert cast(float_value) == variable.static_assign(A=float_value)

    default_var = scalar.AssignedVariable("A", Decimal("1.0"))
    assert default_var == default_var.static_assign(b=2)
    assert cast(float_value) == default_var.static_assign(A=float_value)

    expr = var("a") + var("b")
    assert cast(float_value) + var("b") == expr.static_assign(a=float_value)
    assert var("a") + cast(float_value) == expr.static_assign(b=float_value)

    expr = var("a") * var("b")
    assert cast(float_value) * var("b") == expr.static_assign(a=float_value)
    assert var("a") * cast(float_value) == expr.static_assign(b=float_value)

    expr = var("a") / var("b")
    assert cast(float_value) / var("b") == expr.static_assign(a=float_value)
    assert var("a") / cast(float_value) == expr.static_assign(b=float_value)

    expr = -var("a")
    assert expr == expr.static_assign(b=float_value)
    assert -cast(float_value) == expr.static_assign(a=float_value)

    a = var("a")
    b = var("b")
    c = var("c")
    expr = scalar.Slice(a, scalar.Interval(b, c))
    assert scalar.Slice(cast(float_value), scalar.Interval(b, c)) == expr.static_assign(
        a=float_value
    )
    assert scalar.Slice(a, scalar.Interval(cast(float_value), c)) == expr.static_assign(
        b=float_value
    )
    assert scalar.Slice(a, scalar.Interval(b, cast(float_value))) == expr.static_assign(
        c=float_value
    )
    assert scalar.Slice(
        cast(float_value), scalar.Interval(cast(float_value), c)
    ) == expr.static_assign(a=float_value, b=float_value)
    assert scalar.Slice(
        cast(float_value), scalar.Interval(b, cast(float_value))
    ) == expr.static_assign(a=float_value, c=float_value)
    assert scalar.Slice(
        a, scalar.Interval(cast(float_value), cast(float_value))
    ) == expr.static_assign(b=float_value, c=float_value)
    assert scalar.Slice(
        cast(float_value), scalar.Interval(cast(float_value), cast(float_value))
    ) == expr.static_assign(a=float_value, b=float_value, c=float_value)

    expr = scalar.Min(set(map(cast, ["a", "b", "c"])))
    assert scalar.Min(set(map(cast, [float_value, "b", "c"]))) == expr.static_assign(
        a=float_value
    )

    expr = scalar.Max(set(map(cast, ["a", "b", "c"])))
    assert scalar.Max(set(map(cast, [float_value, "b", "c"]))) == expr.static_assign(
        a=float_value
    )



from bloqade import start, cast
from bloqade.ir.location import Square, Rectangular
import random
import numpy as np
import os

PROJECT_RELATIVE_PPRINT_TESTS_OUTPUT_PATH = os.path.join(
    os.getcwd(), "tests/data/expected_pprint_output"
)


def test_list_of_locations_pprint():
    random.seed(1337)
    rand_positions = list(
        zip(
            [random.randint(0, 20) for _ in range(10)],
            [random.randint(0, 20) for _ in range(10)],
        )
    )

    list_of_locations_pprint_output_path = os.path.join(
        PROJECT_RELATIVE_PPRINT_TESTS_OUTPUT_PATH, "list_of_locations_pprint_output.txt"
    )
    list_of_locations_pprint_output = open(
        list_of_locations_pprint_output_path, "r"
    ).read()

    assert repr(start.add_positions(rand_positions)) == list_of_locations_pprint_output

    var1 = cast("var1")
    var2 = cast("var2")
    var3 = cast("var3")

    assert repr(start.add_positions([(var1, var2), (var3, 5), (0, 9)])) == (
        "[LocationInfo(position=(var(var1), var(var2)), "
        "filling=<SiteFilling.filled: 1>), "
        "LocationInfo(position=(var(var3), 5), filling=<SiteFilling.filled: 1>), "
        "LocationInfo(position=(0, 9), filling=<SiteFilling.filled: 1>)]"
    )


def test_square_pprint():
    # full
    square_pprint_output_path = os.path.join(
        PROJECT_RELATIVE_PPRINT_TESTS_OUTPUT_PATH, "square_pprint_output.txt"
    )
    square_pprint_output = open(square_pprint_output_path, "r").read()

    assert repr(Square(7)) == square_pprint_output

    # apply defect count
    square_pprint_defect_count_output_path = os.path.join(
        PROJECT_RELATIVE_PPRINT_TESTS_OUTPUT_PATH,
        "square_pprint_defect_count_output.txt",
    )
    square_pprint_defect_count_output = open(
        square_pprint_defect_count_output_path, "r"
    ).read()
    assert (
        repr(Square(7).apply_defect_count(21, np.random.default_rng(1337)))
        == square_pprint_defect_count_output
    )

    # apply defect density
    square_pprint_defect_density_output_path = os.path.join(
        PROJECT_RELATIVE_PPRINT_TESTS_OUTPUT_PATH,
        "square_pprint_defect_density_output.txt",
    )
    square_pprint_defect_density_output = open(
        square_pprint_defect_density_output_path, "r"
    ).read()

    assert (
        repr(Square(7).apply_defect_density(0.5, np.random.default_rng(1337)))
        == square_pprint_defect_density_output
    )

    # apply variable for lattice
    square_pprint_var_output_path = os.path.join(
        PROJECT_RELATIVE_PPRINT_TESTS_OUTPUT_PATH, "square_pprint_var_output.txt"
    )
    square_pprint_var_output = open(square_pprint_var_output_path, "r").read()

    bl = cast("bl")
    assert repr(Square(7, bl)) == square_pprint_var_output


def test_rectangular_pprint():
    rectangular_pprint_output_path = os.path.join(
        PROJECT_RELATIVE_PPRINT_TESTS_OUTPUT_PATH, "rectangular_pprint_output.txt"
    )
    rectangular_pprint_output = open(rectangular_pprint_output_path, "r").read()

    assert repr(Rectangular(7, 5)) == rectangular_pprint_output

    rectangular_pprint_defect_count_output_path = os.path.join(
        PROJECT_RELATIVE_PPRINT_TESTS_OUTPUT_PATH,
        "rectangular_pprint_defect_count_output.txt",
    )
    rectangular_pprint_defect_count_output = open(
        rectangular_pprint_defect_count_output_path, "r"
    ).read()

    assert (
        repr(Rectangular(7, 5).apply_defect_count(15, np.random.default_rng(1337)))
        == rectangular_pprint_defect_count_output
    )

    rectangular_pprint_defect_density_output_path = os.path.join(
        PROJECT_RELATIVE_PPRINT_TESTS_OUTPUT_PATH,
        "rectangular_pprint_defect_density_output.txt",
    )
    rectangular_pprint_defect_density_output = open(
        rectangular_pprint_defect_density_output_path, "r"
    ).read()

    assert (
        repr(Rectangular(7, 5).apply_defect_density(0.5, np.random.default_rng(1337)))
        == rectangular_pprint_defect_density_output
    )

    x_spacing = cast("x_spacing")
    assert repr(Rectangular(7, 5, x_spacing)) == (
        "Rectangular(shape=(7, 5), "
        "lattice_spacing=var(x_spacing), "
        "ratio=(1.0 / var(x_spacing)))"
    )



from bloqade.ir.control.waveform import Waveform
from bloqade.ir.control.field import Field, SpatialModulation
from bloqade.ir.control.pulse import PulseExpr
from bloqade.ir.control.sequence import SequenceExpr
from bloqade.ir.location.base import AtomArrangement, ParallelRegister
from bloqade.ir.analog_circuit import AnalogCircuit
from typing import Union, Any


AstType = Union[
    Waveform,
    Field,
    SpatialModulation,
    PulseExpr,
    SequenceExpr,
    AtomArrangement,
    ParallelRegister,
    AnalogCircuit,
]


class AnalogCircuitVisitor:
    def visit_waveform(self, ast: Waveform) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Waveform"
        )

    def visit_field(self, ast: Field) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for Field"
        )

    def visit_spatial_modulation(self, ast: SpatialModulation) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for SpatialModulation"
        )

    def visit_pulse(self, ast: PulseExpr) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for PulseExpr"
        )

    def visit_sequence(self, ast: SequenceExpr) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for SequenceExpr"
        )

    def visit_register(self, ast: AtomArrangement) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for AtomArrangement"
        )

    def visit_parallel_register(self, ast: ParallelRegister) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for MultuplexRegister"
        )

    def visit_analog_circuit(self, ast: AnalogCircuit) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} for AnalogCircuit"
        )

    def visit(self, ast: AstType) -> Any:
        if isinstance(ast, Waveform):
            return self.visit_waveform(ast)
        elif isinstance(ast, SpatialModulation):
            return self.visit_spatial_modulation(ast)
        elif isinstance(ast, Field):
            return self.visit_field(ast)
        elif isinstance(ast, PulseExpr):
            return self.visit_pulse(ast)
        elif isinstance(ast, SequenceExpr):
            return self.visit_sequence(ast)
        elif isinstance(ast, AtomArrangement):
            return self.visit_register(ast)
        elif isinstance(ast, ParallelRegister):
            return self.visit_parallel_register(ast)
        elif isinstance(ast, AnalogCircuit):
            return self.visit_analog_circuit(ast)
        else:
            raise NotImplementedError(f"{ast.__class__} is not a bloqade AST type")



from pydantic import BaseModel
from typing import Optional, List, Tuple
from decimal import Decimal

from bokeh.plotting import figure, show
from bokeh.models import ColumnDataSource, Step, HoverTool
from bokeh.layouts import gridplot

from bloqade.submission.ir.capabilities import QuEraCapabilities
from bloqade.ir.location import ListOfLocations
from bokeh.layouts import row

__all__ = ["QuEraTaskSpecification"]

# TODO: add version to these models.

FloatType = Decimal


def discretize_list(list_of_values: list, resolution: FloatType):
    resolution = Decimal(str(float(resolution)))
    return [round(Decimal(value) / resolution) * resolution for value in list_of_values]


class GlobalField(BaseModel):
    times: List[FloatType]
    values: List[FloatType]

    def __hash__(self):
        return hash((GlobalField, tuple(self.times), tuple(self.values)))


class LocalField(BaseModel):
    times: List[FloatType]
    values: List[FloatType]
    lattice_site_coefficients: List[FloatType]

    def __hash__(self):
        return hash(
            (
                LocalField,
                tuple(self.times),
                tuple(self.values),
                tuple(self.lattice_site_coefficients),
            )
        )


class RabiFrequencyAmplitude(BaseModel):
    global_: GlobalField

    class Config:
        allow_population_by_field_name = True
        fields = {"global_": "global"}

    def __hash__(self):
        return hash((RabiFrequencyAmplitude, self.global_))

    def discretize(self, task_capabilities: QuEraCapabilities):
        global_time_resolution = (
            task_capabilities.capabilities.rydberg.global_.time_resolution
        )
        global_value_resolution = (
            task_capabilities.capabilities.rydberg.global_.rabi_frequency_resolution
        )

        return RabiFrequencyAmplitude(
            global_=GlobalField(
                times=discretize_list(self.global_.times, global_time_resolution),
                values=discretize_list(self.global_.values, global_value_resolution),
            )
        )

    def _get_data_source(self):
        # isolate this for binding glyph later
        src = {
            "times_amp": [float(i) for i in self.global_.times],
            "values_amp": [float(i) for i in self.global_.values],
            "values_floor_amp": [0.0] * len(self.global_.values),
        }

        return src

    def figure(self, source, **fig_kwargs):
        hover = HoverTool()
        hover.tooltips = [("(x,y)", "(@times_amp, @values_amp)")]

        line_plt = figure(
            **fig_kwargs,
            x_axis_label="Time (s)",
            y_axis_label="Ω(t) (rad/s)",
        )

        line_plt.x_range.start = 0
        line_plt.y_range.start = min(source.data["values_amp"]) - 5e6
        line_plt.y_range.end = max(source.data["values_amp"]) + 5e6

        line_plt.line(
            x="times_amp", y="values_amp", source=source, line_width=2, color="black"
        )

        line_plt.circle(
            x="times_amp", y="values_amp", source=source, size=4, color="black"
        )

        line_plt.varea(
            x="times_amp",
            y1="values_amp",
            y2="values_floor_amp",
            source=source,
            fill_alpha=0.3,
            color="#6437FF",
        )
        line_plt.add_tools(hover)

        return line_plt

    def show(self):
        show(self.figure(ColumnDataSource(self._get_data_source())))


class RabiFrequencyPhase(BaseModel):
    global_: GlobalField

    class Config:
        allow_population_by_field_name = True
        fields = {"global_": "global"}

    def __hash__(self):
        return hash((RabiFrequencyPhase, self.global_))

    def discretize(self, task_capabilities: QuEraCapabilities):
        global_time_resolution = (
            task_capabilities.capabilities.rydberg.global_.time_resolution
        )
        global_value_resolution = (
            task_capabilities.capabilities.rydberg.global_.phase_resolution
        )

        return RabiFrequencyPhase(
            global_=GlobalField(
                times=discretize_list(self.global_.times, global_time_resolution),
                values=discretize_list(self.global_.values, global_value_resolution),
            )
        )

    def _get_data_source(self):
        # isolate this for binding glyph later
        src = {
            "times_phase": [float(i) for i in self.global_.times],
            "values_phase": [float(i) for i in self.global_.values],
        }

        return src

    def figure(self, source, **fig_kwargs):
        TOOLTIPS = [("(x,y)", "(@times_phase, @values_phase)")]

        line_plt = figure(
            **fig_kwargs,
            tooltips=TOOLTIPS,
            x_axis_label="Time (s)",
            y_axis_label="ϕ(t) (rad)",
        )

        line_plt.y_range.start = min(source.data["values_phase"]) - 5e6
        line_plt.y_range.end = max(source.data["values_phase"]) + 5e6
        line_plt.x_range.start = 0

        steps = Step(
            x="times_phase",
            y="values_phase",
            line_color="black",
            line_width=2,
            mode="center",
        )

        line_plt.add_glyph(source, steps)

        line_plt.circle(
            x="times_phase", y="values_phase", source=source, size=4, color="black"
        )

        return line_plt

    def show(self):
        show(self.figure(ColumnDataSource(self._get_data_source())))


class Detuning(BaseModel):
    global_: GlobalField
    local: Optional[LocalField]

    class Config:
        allow_population_by_field_name = True
        fields = {"global_": "global"}

    def __hash__(self):
        return hash((Detuning, self.global_, self.local))

    def discretize(self, task_capabilities: QuEraCapabilities):
        global_time_resolution = (
            task_capabilities.capabilities.rydberg.global_.time_resolution
        )
        global_value_resolution = (
            task_capabilities.capabilities.rydberg.global_.detuning_resolution
        )
        local_time_resolution = (
            task_capabilities.capabilities.rydberg.local.time_resolution
        )

        if self.local is not None:
            self.local = LocalField(
                times=discretize_list(self.local.times, local_time_resolution),
                values=self.local.values,
                lattice_site_coefficients=self.local.lattice_site_coefficients,
            )

        return Detuning(
            global_=GlobalField(
                times=discretize_list(self.global_.times, global_time_resolution),
                values=discretize_list(self.global_.values, global_value_resolution),
            ),
            local=self.local,
        )

    def _get_data_source(self):
        # isolate this for binding glyph later
        src = {
            "times_detune": [float(i) for i in self.global_.times],
            "values_detune": [float(i) for i in self.global_.values],
            "values_floor_detune": [0.0] * len(self.global_.values),
        }

        return src

    def global_figure(self, source, **fig_kwargs):
        TOOLTIPS = [("(x,y)", "(@times_detune, @values_detune)")]

        line_plt = figure(
            **fig_kwargs,
            tooltips=TOOLTIPS,
            x_axis_label="Time (s)",
            y_axis_label="Δ(t) (rad/s)",
        )

        line_plt.x_range.start = 0

        line_plt.y_range.start = min(source.data["values_detune"]) - 5e6
        line_plt.y_range.end = max(source.data["values_detune"]) + 5e6

        line_plt.line(
            x="times_detune",
            y="values_detune",
            source=source,
            line_width=2,
            color="black",
        )

        line_plt.circle(
            x="times_detune", y="values_detune", source=source, size=4, color="black"
        )

        line_plt.varea(
            x="times_detune",
            y1="values_detune",
            y2="values_floor_detune",
            source=source,
            fill_alpha=0.5,
            color="#EFD0DE",
        )

        return line_plt

    def show_global(self):
        show(self.global_figure(ColumnDataSource(self._get_data_source())))


class RydbergHamiltonian(BaseModel):
    rabi_frequency_amplitude: RabiFrequencyAmplitude
    rabi_frequency_phase: RabiFrequencyPhase
    detuning: Detuning

    def __hash__(self):
        return hash(
            (
                RydbergHamiltonian,
                self.rabi_frequency_amplitude,
                self.rabi_frequency_phase,
                self.detuning,
            )
        )

    def discretize(self, task_capabilities: QuEraCapabilities):
        return RydbergHamiltonian(
            rabi_frequency_amplitude=self.rabi_frequency_amplitude.discretize(
                task_capabilities
            ),
            rabi_frequency_phase=self.rabi_frequency_phase.discretize(
                task_capabilities
            ),
            detuning=self.detuning.discretize(task_capabilities),
        )


class EffectiveHamiltonian(BaseModel):
    rydberg: RydbergHamiltonian

    def __hash__(self):
        return hash((EffectiveHamiltonian, self.rydberg))

    def discretize(self, task_capabilities: QuEraCapabilities):
        return EffectiveHamiltonian(rydberg=self.rydberg.discretize(task_capabilities))


class Lattice(BaseModel):
    sites: List[Tuple[FloatType, FloatType]]
    filling: List[int]

    def __hash__(self):
        return hash((Lattice, tuple(self.sites), tuple(self.filling)))

    def discretize(self, task_capabilities: QuEraCapabilities):
        position_resolution = (
            task_capabilities.capabilities.lattice.geometry.position_resolution
        )
        return Lattice(
            sites=[discretize_list(site, position_resolution) for site in self.sites],
            filling=self.filling,
        )

    def figure(self, **fig_kwargs):
        ## use ir.Atom_oarrangement's plotting:
        ## covert unit to m -> um
        sites_um = list(
            map(lambda cord: (float(cord[0]) * 1e6, float(cord[1]) * 1e6), self.sites)
        )
        reg = ListOfLocations().add_positions(sites_um, self.filling)
        fig_reg = reg.figure(fig_kwargs=fig_kwargs)  # ignore the B-rad widget
        return fig_reg

    def show(self):
        show(self.figure())


class QuEraTaskSpecification(BaseModel):
    nshots: int
    lattice: Lattice
    effective_hamiltonian: EffectiveHamiltonian

    def __hash__(self):
        return hash(
            (
                QuEraTaskSpecification,
                self.nshots,
                self.lattice,
                self.effective_hamiltonian,
            )
        )

    def discretize(self, task_capabilities: QuEraCapabilities):
        return QuEraTaskSpecification(
            nshots=self.nshots,
            lattice=self.lattice.discretize(task_capabilities),
            effective_hamiltonian=self.effective_hamiltonian.discretize(
                task_capabilities
            ),
        )

    def figure(self):
        # grab all the datas and combine them:
        rabi_amp_src = (
            self.effective_hamiltonian.rydberg.rabi_frequency_amplitude._get_data_source()
        )
        rabi_phase_src = (
            self.effective_hamiltonian.rydberg.rabi_frequency_phase._get_data_source()
        )
        global_detuning_src = (
            self.effective_hamiltonian.rydberg.detuning._get_data_source()
        )

        rabi_amp_src = ColumnDataSource(rabi_amp_src)
        rabi_phase_src = ColumnDataSource(rabi_phase_src)
        global_detuning_src = ColumnDataSource(global_detuning_src)

        # grab global figures
        rabi_amplitude = (
            self.effective_hamiltonian.rydberg.rabi_frequency_amplitude.figure(
                rabi_amp_src, tools="wheel_zoom,reset, undo, redo, pan"
            )
        )

        rabi_phase = self.effective_hamiltonian.rydberg.rabi_frequency_phase.figure(
            rabi_phase_src,
            x_range=rabi_amplitude.x_range,
            tools="hover,wheel_zoom,reset, undo, redo, pan",
        )
        global_detuning = self.effective_hamiltonian.rydberg.detuning.global_figure(
            global_detuning_src,
            x_range=rabi_amplitude.x_range,
            tools="hover,wheel_zoom,reset, undo, redo, pan",
        )

        # lattice:
        register = self.lattice.figure(x_axis_label="x (um)", y_axis_label="y (um)")

        col_plt = gridplot(
            [[rabi_amplitude], [global_detuning], [rabi_phase]],
            merge_tools=False,
            sizing_mode="stretch_both",
        )
        col_plt.width_policy = "max"

        full_plt = row(col_plt, register, sizing_mode="stretch_both")
        full_plt.width_policy = "max"

        return full_plt

    def show(self):
        show(self.figure())



from bloqade.ir import var, cast, Variable, Literal, start


from bloqade.builder.factory import (
    piecewise_linear,
    piecewise_constant,
    linear,
    constant,
)
import bloqade.ir as _ir
from bloqade.builder.parse.json import save_program, load_program
from bloqade.task.json import save_batch, load_batch
from bloqade.constants import RB_C6


def tree_depth(depth: int = None):
    """Setting globally maximum depth for tree printing

    If `depth=None`, return current depth.
    If `depth` is provided, setting current depth to `depth`

    Args:
        depth (int, optional): the user specified depth. Defaults to None.

    Returns:
        int: current updated depth
    """
    if depth is not None:
        _ir.tree_print.max_tree_depth = depth
    return _ir.tree_print.max_tree_depth


__all__ = [
    "RB_C6",
    "start",
    "var",
    "cast",
    "Variable",
    "Literal",
    "piecewise_linear",
    "piecewise_constant",
    "linear",
    "constant",
    "set_print_depth",
    "save_batch",
    "load_batch",
    "save_program",
    "load_program",
]



from bloqade.builder.backend.quera import QuEraService
from bloqade.builder.backend.braket import BraketService
from bloqade.builder.backend.bloqade import BloqadeService


class BackendRoute(QuEraService, BraketService, BloqadeService):
    def device(self, name: str, *args, **kwargs):
        if name == "quera.aquila":
            dev = self.quera.aquila
        elif name == "braket.aquila":
            dev = self.braket.aquila
        elif name == "braket.local_emulator":
            dev = self.braket.local_emulator
        elif name == "bloqade.python":
            dev = self.bloqade.python
        elif name == "bloqade.julia":
            dev = self.bloqade.julia
        else:
            raise ValueError(f"Unknown device: {name}")
        return dev(*args, **kwargs)



from collections import OrderedDict
from dataclasses import dataclass
from typing import Tuple
from numbers import Real

from bloqade.ir.routine.base import RoutineBase
from bloqade.submission.braket import BraketBackend
from bloqade.task.batch import LocalBatch, RemoteBatch
from bloqade.task.braket_simulator import BraketEmulatorTask
from bloqade.task.braket import BraketTask


@dataclass(frozen=True)
class BraketServiceOptions(RoutineBase):
    def aquila(self) -> "BraketHardwareRoutine":
        backend = BraketBackend(
            device_arn="arn:aws:braket:us-east-1::device/qpu/quera/Aquila"
        )
        return BraketHardwareRoutine(source=self.source, backend=backend)

    def local_emulator(self):
        return BraketLocalEmulatorRoutine(source=self.source)


@dataclass(frozen=True)
class BraketHardwareRoutine(RoutineBase):
    backend: BraketBackend

    def compile(
        self,
        shots: int,
        args: Tuple[Real, ...] = (),
        name: str | None = None,
    ) -> RemoteBatch:
        """
        Compile to a RemoteBatch, which contain
            Braket backend specific tasks.

        Args:
            shots (int): number of shots
            args (Tuple): additional arguments
            name (str): custom name of the batch

        Return:
            RemoteBatch

        """

        ## fall passes here ###
        from bloqade.codegen.common.assign_variables import AssignAnalogCircuit
        from bloqade.codegen.common.assignment_scan import AssignmentScan
        from bloqade.codegen.hardware.quera import QuEraCodeGen

        capabilities = self.backend.get_capabilities()

        circuit, params = self.parse_source()
        circuit = AssignAnalogCircuit(params.static_params).visit(circuit)

        tasks = OrderedDict()

        for task_number, batch_params in enumerate(params.batch_assignments(*args)):
            record_params = AssignmentScan(batch_params).emit(circuit)
            final_circuit = AssignAnalogCircuit(record_params).visit(circuit)
            task_ir, parallel_decoder = QuEraCodeGen(capabilities=capabilities).emit(
                shots, final_circuit
            )

            task_ir = task_ir.discretize(capabilities)
            tasks[task_number] = BraketTask(
                None,
                self.backend,
                task_ir,
                batch_params,
                parallel_decoder,
                None,
            )

        batch = RemoteBatch(source=self.source, tasks=tasks, name=name)

        return batch

    def submit(
        self,
        shots: int,
        args: Tuple[Real, ...] = (),
        name: str | None = None,
        shuffle: bool = False,
        **kwargs,
    ) -> RemoteBatch:
        """
        Compile to a RemoteBatch, which contain
        Braket backend specific tasks, and submit to Braket.

        Note:
            This is async.

        Args:
            shots (int): number of shots
            args (Tuple): additional arguments
            name (str): custom name of the batch
            shuffle (bool): shuffle the order of jobs

        Return:
            RemoteBatch

        """

        batch = self.compile(shots, args, name)
        batch._submit(shuffle, **kwargs)
        return batch

    def run(
        self,
        shots: int,
        args: Tuple[Real, ...] = (),
        name: str | None = None,
        shuffle: bool = False,
        **kwargs,
    ) -> RemoteBatch:
        """
        Compile to a RemoteBatch, which contain
        Braket backend specific tasks, submit to Braket,
        and wait until the results are coming back.

        Note:
            This is sync, and will wait until remote results
            finished.

        Args:
            shots (int): number of shots
            args (Tuple): additional arguments
            name (str): custom name of the batch
            shuffle (bool): shuffle the order of jobs

        Return:
            RemoteBatch

        """

        batch = self.submit(shots, args, name, shuffle, **kwargs)
        batch.pull()
        return batch

    def __call__(
        self,
        *args: Tuple[Real, ...],
        shots: int = 1,
        name: str | None = None,
        shuffle: bool = False,
        **kwargs,
    ):
        """
        Compile to a RemoteBatch, which contain
        Braket backend specific tasks, submit to Braket,
        and wait until the results are coming back.

        Note:
            This is sync, and will wait until remote results
            finished.

        Args:
            shots (int): number of shots
            args: additional arguments for flatten variables.
            name (str): custom name of the batch
            shuffle (bool): shuffle the order of jobs

        Return:
            RemoteBatch

        """
        return self.run(shots, args, name, shuffle, **kwargs)


@dataclass(frozen=True)
class BraketLocalEmulatorRoutine(RoutineBase):
    def compile(
        self, shots: int, args: Tuple[Real, ...] = (), name: str | None = None
    ) -> LocalBatch:
        """
        Compile to a LocalBatch, which contain tasks to run on local emulator.

        Args:
            shots (int): number of shots
            args: additional arguments for flatten variables.
            name (str): custom name for the batch

        Return:
            LocalBatch

        """
        ## fall passes here ###
        from bloqade.ir import ParallelRegister
        from bloqade.codegen.common.assign_variables import AssignAnalogCircuit
        from bloqade.codegen.hardware.quera import QuEraCodeGen
        from bloqade.codegen.common.assignment_scan import AssignmentScan
        from bloqade.submission.ir.braket import to_braket_task_ir

        circuit, params = self.parse_source()
        circuit = AssignAnalogCircuit(params.static_params).visit(circuit)

        if isinstance(circuit.register, ParallelRegister):
            raise TypeError(
                "Parallelization of atom arrangements is not supported for "
                "local emulation."
            )

        tasks = OrderedDict()

        for task_number, batch_params in enumerate(params.batch_assignments(*args)):
            record_params = AssignmentScan(batch_params).emit(circuit)
            final_circuit = AssignAnalogCircuit(record_params).visit(circuit)
            quera_task_ir, _ = QuEraCodeGen().emit(shots, final_circuit)

            task_ir = to_braket_task_ir(quera_task_ir)

            tasks[task_number] = BraketEmulatorTask(
                task_ir,
                batch_params,
                None,
            )

        batch = LocalBatch(source=self.source, tasks=tasks, name=name)

        return batch

    def run(
        self,
        shots: int,
        args: Tuple[Real, ...] = (),
        name: str | None = None,
        multiprocessing: bool = False,
        num_workers: int | None = None,
        **kwargs,
    ) -> LocalBatch:
        """
        Compile to a LocalBatch, and run.
        The LocalBatch contain tasks to run on local emulator.

        Note:
            This is sync, and will wait until remote results
            finished.

        Args:
            shots (int): number of shots
            args: additional arguments for flatten variables.
            multiprocessing (bool): enable multi-process
            num_workers (int): number of workers to run the emulator

        Return:
            LocalBatch

        """

        batch = self.compile(shots, args, name)
        batch._run(multiprocessing=multiprocessing, num_workers=num_workers, **kwargs)
        return batch



from bloqade.builder.base import Builder
from typing import Any, List, TYPE_CHECKING

if TYPE_CHECKING:
    from bloqade.builder.parallelize import Parallelize
    from bloqade.builder.assign import Assign, BatchAssign
    from bloqade.builder.flatten import Flatten


class Flattenable(Builder):
    def flatten(self, orders: List[str]) -> "Flatten":
        from bloqade.builder.flatten import Flatten

        return Flatten(orders, self)


class Assignable(Builder):
    def assign(self, **assignments) -> "Assign":
        """
        Assign values to variables declared previously in the program.

        Args:
            assignments (Dict[str, Union[Number]]):
            The assignments, which should be a kwargs
            where the key is the variable name and the
            value is the value to assign to the variable.

        Examples:
            - Assign the value 0.0 to the variable "ival"
            and 0.5 to the variable "span_time".

            >>> reg = bloqade.start
            ...       .add_positions([(0,0),(1,1),(2,2),(3,3)])
            >>> seq = reg.rydberg.detuning.uniform
            ...       .linear(start="ival",stop=1,duration="span_time")
            >>> seq = seq.assign(span_time = 0.5, ival = 0.0)

        """
        from bloqade.builder.assign import Assign

        return Assign(parent=self, **assignments)


class BatchAssignable(Builder):
    def batch_assign(self, **assignments) -> "BatchAssign":
        from bloqade.builder.assign import BatchAssign

        return BatchAssign(parent=self, **assignments)


class Parallelizable(Builder):
    def parallelize(self, cluster_spacing: Any) -> "Parallelize":
        """
        Parallelize the current problem (register & sequnece) to fill entire FOV
        with the given cluster spacing.

        Args:
            cluster_spacing (Any, should be Real):
            the spacing between parallel clusters.

        Examples:
            - Parallelize the current problem with cluster spacing 7.2 um.

            >>> prob = (
                    bloqade.start.add_positions([(0,0),(1,1),(2,2),(3,3)])
                    .rydberg.detuning.uniform
                    .linear(start=0,stop=1,duration=1)
                    )
            >>> prob = prob.parallelize(7.2)

        """
        from bloqade.builder.parallelize import Parallelize

        return Parallelize(cluster_spacing, self)



from bloqade.builder.start import ProgramStart
from bloqade.ir.scalar import Scalar, Literal, cast
from pydantic.dataclasses import dataclass
from typing import List, Generator, Tuple, Optional, Any, TYPE_CHECKING
from bokeh.plotting import show
import numpy as np
from enum import Enum
from bokeh.models import (
    ColumnDataSource,
    NumericInput,
    Button,
    Range1d,
    CustomJS,
    HoverTool,
)
from bokeh.plotting import figure
from bokeh.layouts import column, row
import plotext as pltxt
import sys

# from bokeh import events


if TYPE_CHECKING:
    from .list import ListOfLocations


class SiteFilling(int, Enum):
    filled = 1
    vacant = 0


@dataclass
class LocationInfo:
    position: Tuple[Scalar, Scalar]
    filling: SiteFilling

    def __init__(self, position: Tuple[Any, Any], filled: bool):
        if filled:
            self.filling = SiteFilling.filled
        else:
            self.filling = SiteFilling.vacant

        self.position = tuple(cast(ele) for ele in position)


class AtomArrangement(ProgramStart):
    def __repr__(self) -> str:
        xs_filled, xs_vacant = [], []
        ys_filled, ys_vacant = [], []

        counter = 0
        for _, location_info in enumerate(self.enumerate()):
            counter += 1
            (x, y) = location_info.position
            if type(x) is not Literal or type(y) is not Literal:
                return repr(
                    list(self.enumerate())
                )  # default to standard print of internal contents
            else:
                if location_info.filling is SiteFilling.filled:
                    xs_filled.append(float(x.value))
                    ys_filled.append(float(y.value))
                else:
                    xs_vacant.append(float(x.value))
                    ys_vacant.append(float(y.value))

        if counter == 0:
            return repr(
                list(self.enumerate())
            )  # default to standard print of internal contents

        pltxt.clear_figure()
        pltxt.limit_size(False, False)
        pltxt.plot_size(80, 24)
        pltxt.canvas_color("default")
        pltxt.axes_color("default")
        pltxt.ticks_color("white")
        pltxt.title("Atom Positions")
        pltxt.xlabel("x (um)")
        pltxt.ylabel("y (um)")

        marker = "."
        if sys.stdout.encoding.lower().startswith("utf"):
            marker = "dot"

        pltxt.scatter(xs_filled, ys_filled, color=(100, 55, 255), marker=marker)
        if len(xs_vacant) > 0:
            pltxt.scatter(
                xs_vacant, ys_vacant, color="white", label="vacant", marker=marker
            )

        return pltxt.build()

    def enumerate(self) -> Generator[LocationInfo, None, None]:
        """enumerate all locations in the register."""
        raise NotImplementedError

    def figure(self, fig_kwargs=None, **assignments):
        """obtain a figure object from the atom arrangement."""
        xs_filled, ys_filled, labels_filled = [], [], []
        xs_vacant, ys_vacant, labels_vacant = [], [], []
        x_min = np.inf
        x_max = -np.inf
        y_min = np.inf
        y_max = -np.inf
        for idx, location_info in enumerate(self.enumerate()):
            (x_var, y_var) = location_info.position
            (x, y) = (float(x_var(**assignments)), float(y_var(**assignments)))
            x_min = min(x, x_min)
            y_min = min(y, y_min)
            x_max = max(x, x_max)
            y_max = max(y, y_max)
            if location_info.filling is SiteFilling.filled:
                xs_filled.append(x)
                ys_filled.append(y)
                labels_filled.append(idx)
            else:
                xs_vacant.append(x)
                ys_vacant.append(y)
                labels_vacant.append(idx)

        # Ly = y_max - y_min
        # Lx = x_max - x_min
        # scale_x = (Lx+2)/(Ly+2)

        if self.n_atoms > 0:
            length_scale = max(y_max - y_min, x_max - x_min, 1)
        else:
            length_scale = 1

        source_filled = ColumnDataSource(
            data=dict(_x=xs_filled, _y=ys_filled, _labels=labels_filled)
        )
        source_vacant = ColumnDataSource(
            data=dict(_x=xs_vacant, _y=ys_vacant, _labels=labels_vacant)
        )
        source_all = ColumnDataSource(
            data=dict(
                _x=xs_vacant + xs_filled,
                _y=ys_vacant + ys_filled,
                _labels=labels_vacant + labels_filled,
            )
        )
        hover = HoverTool()
        hover.tooltips = [
            ("(x,y)", "(@_x, @_y)"),
            ("index: ", "@_labels"),
        ]

        ## remove box_zoom since we don't want to change the scale
        if fig_kwargs is None:
            fig_kwargs = {}

        p = figure(
            **fig_kwargs,
            width=400,
            height=400,
            tools="wheel_zoom,reset, undo, redo, pan",
            toolbar_location="above",
        )
        p.x_range = Range1d(x_min - 1, x_min + length_scale + 1)
        p.y_range = Range1d(y_min - 1, y_min + length_scale + 1)

        p.circle(
            "_x", "_y", source=source_filled, radius=0.015 * length_scale, fill_alpha=1
        )
        p.circle(
            "_x",
            "_y",
            source=source_vacant,
            radius=0.015 * length_scale,
            fill_alpha=0.25,
            color="grey",
            line_width=0.2 * length_scale,
        )

        p.circle(
            "_x",
            "_y",
            source=source_all,
            radius=0,  # in the same unit as the data
            fill_alpha=0,
            line_width=0.15 * length_scale,
            visible=True,  # display by default
            name="Brad",
        )
        p.add_tools(hover)

        return p

    def show(self, **assignments) -> None:
        """show the register."""
        p = self.figure(**assignments)

        # get the Blocade rad object
        cr = None
        for rd in p.renderers:
            if rd.name == "Brad":
                cr = rd

        # adding rydberg radis input
        Brad_input = NumericInput(
            value=0, low=0, title="Bloqade radius (um):", mode="float"
        )

        # js link toggle btn
        toggle_button = Button(label="Toggle")
        toggle_button.js_on_event(
            "button_click",
            CustomJS(args=dict(cr=cr), code="""cr.visible = !cr.visible;"""),
        )

        # js link radius
        Brad_input.js_link("value", cr.glyph, "radius")

        full = column(p, row(Brad_input, toggle_button))
        # full.sizing_mode="scale_both"
        show(full)

    @property
    def n_atoms(self) -> int:
        """number of atoms in the register."""
        raise NotImplementedError

    @property
    def n_dims(self) -> int:
        """number of dimensions in the register."""
        raise NotImplementedError

    def scale(self, scale: float | Scalar) -> "ListOfLocations":
        """scale the atom arrangement with a given factor"""
        from .list import ListOfLocations

        scale = cast(scale)
        location_list = []
        for location_info in self.enumerate():
            x, y = location_info.position
            new_position = (scale * x, scale * y)
            location_list.append(
                LocationInfo(new_position, bool(location_info.filling.value))
            )

        return ListOfLocations(location_list)

    def add_position(
        self, position: Tuple[Any, Any], filled: bool = True
    ) -> "ListOfLocations":
        """add a position to existing atom arrangement."""

        from .list import ListOfLocations

        location_list = [LocationInfo(position, filled)]
        for location_info in self.enumerate():
            location_list.append(location_info)

        return ListOfLocations(location_list)

    def add_positions(
        self, positions: List[Tuple[Any, Any]], filling: Optional[List[bool]] = None
    ) -> "ListOfLocations":
        """add a list of positions to existing atom arrangement."""
        from .list import ListOfLocations

        location_list = []

        if filling:
            for position, filled in zip(positions, filling):
                location_list.append(LocationInfo(position, filled))

        else:
            for position in positions:
                location_list.append(LocationInfo(position, True))

        for location_info in self.enumerate():
            location_list.append(location_info)

        return ListOfLocations(location_list)

    def apply_defect_count(
        self, n_defects: int, rng: np.random.Generator = np.random.default_rng()
    ) -> "ListOfLocations":
        """apply n_defects randomly to existing atom arrangement."""
        from .list import ListOfLocations

        location_list = []
        for location_info in self.enumerate():
            location_list.append(location_info)

        for _ in range(n_defects):
            idx = rng.integers(0, len(location_list))
            location_list[idx] = LocationInfo(
                location_list[idx].position,
                (False if location_list[idx].filling is SiteFilling.filled else True),
            )

        return ListOfLocations(location_list)

    def apply_defect_density(
        self,
        defect_probability: float,
        rng: np.random.Generator = np.random.default_rng(),
    ) -> "ListOfLocations":
        """apply defect_probability randomly to existing atom arrangement."""
        from .list import ListOfLocations

        p = min(1, max(0, defect_probability))
        location_list = []

        for location_info in self.enumerate():
            if rng.random() < p:
                location_list.append(
                    LocationInfo(
                        location_info.position,
                        (
                            False
                            if location_info.filling is SiteFilling.filled
                            else True
                        ),
                    )
                )
            else:
                location_list.append(location_info)

        return ListOfLocations(location_list=location_list)


@dataclass(init=False)
class ParallelRegister(ProgramStart):
    """Parallel Register"""

    __match_args__ = ("_register", "_cluster_spacing")

    register_locations: List[List[Scalar]]
    register_filling: List[int]
    shift_vectors: List[List[Scalar]]

    def __init__(self, register: AtomArrangement, cluster_spacing: Any):
        self._register = register
        self._cluster_spacing = cluster_spacing

        if register.n_atoms > 0:
            # calculate bounding box
            # of this register
            location_iter = register.enumerate()
            (x, y) = next(location_iter).position
            x_min = x
            x_max = x
            y_min = y
            y_max = y

            for location_info in location_iter:
                (x, y) = location_info.position
                x_min = x.min(x_min)
                x_max = x.max(x_max)
                y_min = y.min(y_min)
                y_max = y.max(y_max)

            shift_x = (x_max - x_min) + cluster_spacing
            shift_y = (y_max - y_min) + cluster_spacing

            register_locations = [
                list(location_info.position) for location_info in register.enumerate()
            ]
            register_filling = [
                location_info.filling.value for location_info in register.enumerate()
            ]
            shift_vectors = [[shift_x, cast(0)], [cast(0), shift_y]]
        else:
            raise ValueError("No locations to parallelize.")

        self.register_locations = register_locations
        self.register_filling = register_filling
        self.shift_vectors = shift_vectors
        super().__init__(self)



from bloqade.builder.base import Builder


class LevelCoupling(Builder):
    @property
    def detuning(self):
        """
        - Specify the Detuning field
        - Next-step: <SpacialModulation>
        - Possible Next:

            -> `...detuning.location(int)`
                :: Address atom at specific location

            -> `...detuning.uniform`
                :: Address all atoms in register

            -> `...detuning.var(str)`
                :: Address atom at location labeled by variable

        """
        from bloqade.builder.field import Detuning

        return Detuning(self)

    @property
    def rabi(self):
        """
        - Specify the Rabi term/field.
        - Possible Next:

            -> `...rabi.amplitude`
                :: address rabi amplitude

            -> `...rabi.phase`
                :: address rabi phase


        """
        from bloqade.builder.field import Rabi

        return Rabi(self)


class Rydberg(LevelCoupling):
    """
    This node represent level coupling of rydberg state.

    Examples:

        - To reach the node from the start node:

        >>> node = bloqade.start.rydberg
        >>> type(node)
        <class 'bloqade.builder.coupling.Rydberg'>

        - Rydberg level coupling have two reachable field nodes:

            - detuning term (See also [`Detuning`][bloqade.builder.field.Detuning])
            - rabi term (See also [`Rabi`][bloqade.builder.field.Rabi])

        >>> ryd_detune = bloqade.start.rydberg.detuning
        >>> ryd_rabi = bloqade.start.rydberg.rabi

    """

    def __bloqade_ir__(self):
        from bloqade.ir.control.sequence import rydberg

        return rydberg


class Hyperfine(LevelCoupling):
    """
    This node represent level coupling between hyperfine state.

    Examples:

        - To reach the node from the start node:

        >>> node = bloqade.start.hyperfine
        >>> type(node)
        <class 'bloqade.builder.coupling.Hyperfine'>

        - Hyperfine level coupling have two reachable field nodes:

            - detuning term (See also [`Detuning`][bloqade.builder.field.Detuning])
            - rabi term (See also [`Rabi`][bloqade.builder.field.Rabi])

        >>> hyp_detune = bloqade.start.hyperfine.detuning
        >>> hyp_rabi = bloqade.start.hyperfine.rabi

    """

    def __bloqade_ir__(self):
        from bloqade.ir.control.sequence import hyperfine

        return hyperfine



from typing import List
from bloqade.builder.base import Builder
from bloqade.builder.backend import BackendRoute


class Flatten(BackendRoute):
    __match_args__ = ("_order", "__parent__")

    def __init__(self, order: List[str], parent: Builder | None = None) -> None:
        super().__init__(parent)
        self._order = tuple(order)



from bloqade.ir import Sequence, rydberg, detuning, Uniform, Linear, ScaledLocations
from bloqade.ir.location import Square

# import bloqade.lattice as lattice

n_atoms = 11
lattice_const = 5.9

rabi_amplitude_values = [0.0, 15.8, 15.8, 0.0]
rabi_detuning_values = [-16.33, -16.33, "delta_end", "delta_end"]
durations = [0.8, "sweep_time", 0.8]

ordered_state_2D_prog = (
    Square(n_atoms, lattice_const)
    .rydberg.rabi.amplitude.uniform.piecewise_linear(durations, rabi_amplitude_values)
    .detuning.uniform.piecewise_linear(durations, rabi_detuning_values)
)

ordered_state_2D_job = ordered_state_2D_prog.assign(delta_end=42.66, sweep_time=2.4)

pbin = ordered_state_2D_job.quera.aquila()

pbin = pbin.parse_circuit()

# pbin.circuit.sequence


# dict interface
seq = Sequence(
    {
        rydberg: {
            detuning: {
                Uniform: Linear(start=1.0, stop="x", duration=3.0),
                ScaledLocations({1: 1.0, 2: 2.0}): Linear(
                    start=1.0, stop="x", duration=3.0
                ),
            },
        }
    }
)


# job = HardwareBatchResult.load_json("example-3-2d-ordered-state-job.json")

# res = job.report()


# print(lattice.Square(3).apply(seq).__lattice__)
# print(lattice.Square(3).apply(seq).braket(nshots=1000).submit().report().dataframe)
# print("bitstring")
# print(lattice.Square(3).apply(seq).braket(nshots=1000).submit().report().bitstring)

# # pipe interface
# report = (
#     lattice.Square(3)
#     .rydberg.detuning.uniform.apply(Linear(start=1.0, stop="x", duration=3.0))
#     .location(2)
#     .scale(3.0)
#     .apply(Linear(start=1.0, stop="x", duration=3.0))
#     .rydberg.rabi.amplitude.uniform
#     .apply(Linear(start=1.0, stop="x", duration=3.0))
#     .assign(x=10)
#     .braket(nshots=1000)
#     .submit()
#     .report()
# )

# print(report)
# print(report.bitstring)
# print(report.dataframe)

# lattice.Square(3).rydberg.detuning.location(2).location(3).apply(
#     Linear(start=1.0, stop="x", duration=3.0)
# ).location(3).location(4).apply(Linear(start=1.0, stop="x", duration=3.0)).braket(
#     nshots=1000
# ).submit()

# # start.rydberg.detuning.location(2).location(3)


# prog = (
#     lattice.Square(3)
#     .rydberg.detuning.uniform.apply(Linear(start=1.0, stop="x", duration=3.0))
#     .location(2)
#     .scale(3.0)
#     .apply(Linear(start=1.0, stop="x", duration=3.0))
#     .hyperfine.rabi.amplitude.location(2)
#     .apply(Linear(start=1.0, stop="x", duration=3.0))
#     .assign(x=1.0)
#     .multiplex(10.0).braket(nshots=1000)
#     .submit()
#     .report()
#     .dataframe.groupby(by=["x"])
#     .count()
# )

# (
#     lattice.Square(3)
#     .rydberg.detuning.uniform.apply(Linear(start=1.0, stop="x", duration=3.0))
#     .multiplex.quera
# )


# wf = (
#     Linear(start=1.0, stop=2.0, duration=2.0)
#     .scale(2.0)
#     .append(Linear(start=1.0, stop=2.0, duration=2.0))
# )


# prog = (
#     lattice.Square(3)
#     .hyperfine.detuning.location(1)
#     .scale(2.0)
#     .piecewise_linear(coeffs=[1.0, 2.0, 3.0])
#     .location(2)
#     .constant(value=2.0, duration="x")
# )

# prog.seq
# prog.lattice



import bloqade.ir.location as location
from bloqade.ir import Linear, Constant

quantum_job = (
    location.Square(6)
    .rydberg.detuning.uniform.apply(
        Constant("initial_detuning", "up_time")
        .append(Linear("initial_detuning", "final_detuning", "anneal_time"))
        .append(Constant("final_detuning", "up_time"))
    )
    .rabi.amplitude.uniform.apply(
        Linear(0.0, "rabi_amplitude_max", "up_time")
        .append(Constant("rabi_amplitude_max", "anneal_time"))
        .append(Linear("rabi_amplitude_max", 0.0, "up_time"))
    )
    .assign(
        initial_detuning=-10,
        up_time=0.1,
        final_detuning=15,
        anneal_time=10,
        rabi_amplitude_max=15,
    )
    .quera.mock()
    .compile(shots=10)
)

# print(len(quantum_task.task_result.shot_outputs))
# job_file = tempfile.mktemp(suffix=".json")

# quantum_job.save_json(job_file)
# quantum_job = HardwareBatchTask.load_json(job_file)

# future_file = tempfile.mktemp(suffix=".json")
# quantum_future = quantum_job.submit()
# quantum_future.save_json(future_file)
# quantum_future = HardwareBatchResult.load_json(future_file)

# quantum_future.json()
# quantum_future.task_results
